# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.core.expr.numvalue import value
from pyomo.solvers.plugins.solvers.cplex_direct import CPLEXDirect
from pyomo.solvers.plugins.solvers.persistent_solver import PersistentSolver
from pyomo.opt.base import SolverFactory


@SolverFactory.register('cplex_persistent', doc='Persistent python interface to CPLEX')
class CPLEXPersistent(PersistentSolver, CPLEXDirect):
    """
    A class that provides a persistent interface to Cplex. Direct solver interfaces do not use any file io.
    Rather, they interface directly with the python bindings for the specific solver. Persistent solver interfaces
    are similar except that they "remember" their model. Thus, persistent solver interfaces allow incremental changes
    to the solver model (e.g., the gurobi python model or the cplex python model). Note that users are responsible
    for notifying the persistent solver interfaces when changes are made to the corresponding pyomo model.

    Keyword Arguments
    -----------------
    model: ConcreteModel
        Passing a model to the constructor is equivalent to calling the set_instance method.
    type: str
        String indicating the class type of the solver instance.
    name: str
        String representing either the class type of the solver instance or an assigned name.
    doc: str
        Documentation for the solver
    options: dict
        Dictionary of solver options
    """

    def __init__(self, **kwds):
        kwds['type'] = 'cplex_persistent'
        PersistentSolver.__init__(self, **kwds)
        CPLEXDirect._init(self)

        self._pyomo_model = kwds.pop('model', None)
        if self._pyomo_model is not None:
            self.set_instance(self._pyomo_model, **kwds)

    def _remove_constraint(self, solver_con):
        try:
            self._solver_model.linear_constraints.delete(solver_con)
        except self._cplex.exceptions.CplexError:
            try:
                self._solver_model.quadratic_constraints.delete(solver_con)
            except self._cplex.exceptions.CplexError:
                raise ValueError(
                    'Failed to find the cplex constraint {0}'.format(solver_con)
                )

    def _remove_sos_constraint(self, solver_sos_con):
        self._solver_model.SOS.delete(solver_sos_con)

    def _remove_var(self, solver_var):
        pyomo_var = self._solver_var_to_pyomo_var_map[solver_var]
        ndx = self._pyomo_var_to_ndx_map[pyomo_var]
        for tmp_var, tmp_ndx in self._pyomo_var_to_ndx_map.items():
            if tmp_ndx > ndx:
                self._pyomo_var_to_ndx_map[tmp_var] -= 1
        self._ndx_count -= 1
        del self._pyomo_var_to_ndx_map[pyomo_var]
        self._solver_model.variables.delete(solver_var)

    def _warm_start(self):
        CPLEXDirect._warm_start(self)

    def update_var(self, var):
        """Update a single variable in the solver's model.

        This will update bounds, fix/unfix the variable as needed, and
        update the variable type.

        Parameters
        ----------
        var: Var (scalar Var or single VarData)

        """
        # see PR #366 for discussion about handling indexed
        # objects and keeping compatibility with the
        # pyomo.kernel objects
        # if var.is_indexed():
        #    for child_var in var.values():
        #        self.compile_var(child_var)
        #    return
        if var not in self._pyomo_var_to_solver_var_map:
            raise ValueError(
                'The Var provided to compile_var needs to be added first: {0}'.format(
                    var
                )
            )
        cplex_var = self._pyomo_var_to_solver_var_map[var]
        vtype = self._cplex_vtype_from_var(var)
        lb, ub = self._cplex_lb_ub_from_var(var)

        self._solver_model.variables.set_lower_bounds(cplex_var, lb)
        self._solver_model.variables.set_upper_bounds(cplex_var, ub)
        self._solver_model.variables.set_types(cplex_var, vtype)

    def write(self, filename, filetype=''):
        """
        Write the model to a file (e.g., and lp file).

        Parameters
        ----------
        filename: str
            Name of the file to which the model should be written.
        filetype: str
            The file type (e.g., lp).
        """
        self._solver_model.write(filename, filetype=filetype)

    def _add_column(self, var, obj_coef, constraints, coefficients):
        """Add a column to the solver's model

        This will add the Pyomo variable var to the solver's
        model, and put the coefficients on the associated
        constraints in the solver model. If the obj_coef is
        not zero, it will add obj_coef*var to the objective
        of the solver's model.

        Parameters
        ----------
        var: Var (scalar Var or single VarData)
        obj_coef: float
        constraints: list of solver constraints
        coefficients: list of coefficients to put on var in the associated constraint
        """

        ## set-up add var
        varname = self._symbol_map.getSymbol(var, self._labeler)
        vtype = self._cplex_vtype_from_var(var)
        lb, ub = self._cplex_lb_ub_from_var(var)

        ## do column addition
        self._solver_model.variables.add(
            obj=[obj_coef],
            lb=[lb],
            ub=[ub],
            types=[vtype],
            names=[varname],
            columns=[self._cplex.SparsePair(ind=constraints, val=coefficients)],
        )

        self._pyomo_var_to_solver_var_map[var] = varname
        self._solver_var_to_pyomo_var_map[varname] = var
        self._pyomo_var_to_ndx_map[var] = self._ndx_count
        self._ndx_count += 1
        self._referenced_variables[var] = len(coefficients)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest
import pyomo.environ as pyo
from pyomo.core.expr.taylor_series import taylor_series_expansion

try:
    import gurobipy

    with gurobipy.Env():
        pass

    gurobipy_available = True
except:
    gurobipy_available = False


@unittest.skipIf(not gurobipy_available, "gurobipy is not available")
@unittest.pytest.mark.solver("gurobi_persistent")
class TestGurobiPersistent(unittest.TestCase):
    def test_basics(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var(bounds=(-10, 10))
        m.y = pyo.Var()
        m.obj = pyo.Objective(expr=m.x**2 + m.y**2)
        m.c1 = pyo.Constraint(expr=m.y >= 2 * m.x + 1)

        opt = pyo.SolverFactory('gurobi_persistent')
        opt.set_instance(m)

        self.assertEqual(opt.get_model_attr('NumVars'), 2)
        self.assertEqual(opt.get_model_attr('NumConstrs'), 1)
        self.assertEqual(opt.get_model_attr('NumQConstrs'), 0)
        self.assertEqual(opt.get_var_attr(m.x, 'LB'), -10)
        self.assertEqual(opt.get_var_attr(m.x, 'UB'), 10)

        res = opt.solve()
        self.assertAlmostEqual(m.x.value, -0.4)
        self.assertAlmostEqual(m.y.value, 0.2)
        opt.load_duals()
        self.assertAlmostEqual(m.dual[m.c1], -0.4)
        del m.dual

        m.c2 = pyo.Constraint(expr=m.y >= -m.x + 1)
        opt.add_constraint(m.c2)
        self.assertEqual(opt.get_model_attr('NumVars'), 2)
        self.assertEqual(opt.get_model_attr('NumConstrs'), 2)
        self.assertEqual(opt.get_model_attr('NumQConstrs'), 0)

        res = opt.solve(save_results=False, load_solutions=False)
        self.assertAlmostEqual(m.x.value, -0.4)
        self.assertAlmostEqual(m.y.value, 0.2)
        opt.load_vars()
        self.assertAlmostEqual(m.x.value, 0)
        self.assertAlmostEqual(m.y.value, 1)

        opt.remove_constraint(m.c2)
        m.del_component(m.c2)
        self.assertEqual(opt.get_model_attr('NumVars'), 2)
        self.assertEqual(opt.get_model_attr('NumConstrs'), 1)
        self.assertEqual(opt.get_model_attr('NumQConstrs'), 0)

        self.assertEqual(opt.get_gurobi_param_info('FeasibilityTol')[2], 1e-6)
        res = opt.solve(options={'FeasibilityTol': '1e-7'})
        self.assertEqual(opt.get_gurobi_param_info('FeasibilityTol')[2], 1e-7)
        self.assertAlmostEqual(m.x.value, -0.4)
        self.assertAlmostEqual(m.y.value, 0.2)

        m.x.setlb(-5)
        m.x.setub(5)
        opt.update_var(m.x)
        self.assertEqual(opt.get_var_attr(m.x, 'LB'), -5)
        self.assertEqual(opt.get_var_attr(m.x, 'UB'), 5)

        m.x.fix(0)
        opt.update_var(m.x)
        self.assertEqual(opt.get_var_attr(m.x, 'LB'), 0)
        self.assertEqual(opt.get_var_attr(m.x, 'UB'), 0)

        m.x.unfix()
        opt.update_var(m.x)
        self.assertEqual(opt.get_var_attr(m.x, 'LB'), -5)
        self.assertEqual(opt.get_var_attr(m.x, 'UB'), 5)

        m.c2 = pyo.Constraint(expr=m.y >= m.x**2)
        opt.add_constraint(m.c2)
        self.assertEqual(opt.get_model_attr('NumVars'), 2)
        self.assertEqual(opt.get_model_attr('NumConstrs'), 1)
        self.assertEqual(opt.get_model_attr('NumQConstrs'), 1)

        opt.remove_constraint(m.c2)
        m.del_component(m.c2)
        self.assertEqual(opt.get_model_attr('NumVars'), 2)
        self.assertEqual(opt.get_model_attr('NumConstrs'), 1)
        self.assertEqual(opt.get_model_attr('NumQConstrs'), 0)

        m.z = pyo.Var()
        opt.add_var(m.z)
        self.assertEqual(opt.get_model_attr('NumVars'), 3)
        opt.remove_var(m.z)
        del m.z
        self.assertEqual(opt.get_model_attr('NumVars'), 2)

    def test_update1(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.z = pyo.Var()
        m.obj = pyo.Objective(expr=m.z)
        m.c1 = pyo.Constraint(expr=m.z >= m.x**2 + m.y**2)

        opt = pyo.SolverFactory('gurobi_persistent')
        opt.set_instance(m)
        self.assertEqual(opt._solver_model.getAttr('NumQConstrs'), 0)

        opt.remove_constraint(m.c1)
        opt.update()
        self.assertEqual(opt._solver_model.getAttr('NumQConstrs'), 0)

        opt.add_constraint(m.c1)
        self.assertEqual(opt._solver_model.getAttr('NumQConstrs'), 0)
        opt.update()
        self.assertEqual(opt._solver_model.getAttr('NumQConstrs'), 1)

    def test_update2(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.z = pyo.Var()
        m.obj = pyo.Objective(expr=m.z)
        m.c2 = pyo.Constraint(expr=m.x + m.y == 1)

        opt = pyo.SolverFactory('gurobi_persistent')
        opt.set_instance(m)
        self.assertEqual(opt._solver_model.getAttr('NumConstrs'), 0)

        opt.remove_constraint(m.c2)
        opt.update()
        self.assertEqual(opt._solver_model.getAttr('NumConstrs'), 0)

        opt.add_constraint(m.c2)
        self.assertEqual(opt._solver_model.getAttr('NumConstrs'), 0)
        opt.update()
        self.assertEqual(opt._solver_model.getAttr('NumConstrs'), 1)

    def test_update3(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.z = pyo.Var()
        m.obj = pyo.Objective(expr=m.z)
        m.c1 = pyo.Constraint(expr=m.z >= m.x**2 + m.y**2)

        opt = pyo.SolverFactory('gurobi_persistent')
        opt.set_instance(m)
        opt.update()
        self.assertEqual(opt._solver_model.getAttr('NumQConstrs'), 1)
        m.c2 = pyo.Constraint(expr=m.y >= m.x**2)
        opt.add_constraint(m.c2)
        self.assertEqual(opt._solver_model.getAttr('NumQConstrs'), 1)
        opt.remove_constraint(m.c2)
        opt.update()
        self.assertEqual(opt._solver_model.getAttr('NumQConstrs'), 1)

    def test_update4(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.z = pyo.Var()
        m.obj = pyo.Objective(expr=m.z)
        m.c1 = pyo.Constraint(expr=m.z >= m.x + m.y)

        opt = pyo.SolverFactory('gurobi_persistent')
        opt.set_instance(m)
        opt.update()
        self.assertEqual(opt._solver_model.getAttr('NumConstrs'), 1)
        m.c2 = pyo.Constraint(expr=m.y >= m.x)
        opt.add_constraint(m.c2)
        self.assertEqual(opt._solver_model.getAttr('NumConstrs'), 1)
        opt.remove_constraint(m.c2)
        opt.update()
        self.assertEqual(opt._solver_model.getAttr('NumConstrs'), 1)

    def test_update5(self):
        m = pyo.ConcreteModel()
        m.a = pyo.Set(initialize=[1, 2, 3], ordered=True)
        m.x = pyo.Var(m.a, within=pyo.Binary)
        m.y = pyo.Var(within=pyo.Binary)
        m.obj = pyo.Objective(expr=m.y)
        m.c1 = pyo.SOSConstraint(var=m.x, sos=1)

        opt = pyo.SolverFactory('gurobi_persistent')
        opt.set_instance(m)
        self.assertEqual(opt._solver_model.getAttr('NumSOS'), 0)

        opt.remove_sos_constraint(m.c1)
        opt.update()
        self.assertEqual(opt._solver_model.getAttr('NumSOS'), 0)

        opt.add_sos_constraint(m.c1)
        self.assertEqual(opt._solver_model.getAttr('NumSOS'), 0)
        opt.update()
        self.assertEqual(opt._solver_model.getAttr('NumSOS'), 1)

    def test_update6(self):
        m = pyo.ConcreteModel()
        m.a = pyo.Set(initialize=[1, 2, 3], ordered=True)
        m.x = pyo.Var(m.a, within=pyo.Binary)
        m.y = pyo.Var(within=pyo.Binary)
        m.obj = pyo.Objective(expr=m.y)
        m.c1 = pyo.SOSConstraint(var=m.x, sos=1)

        opt = pyo.SolverFactory('gurobi_persistent')
        opt.set_instance(m)
        opt.update()
        self.assertEqual(opt._solver_model.getAttr('NumSOS'), 1)
        m.c2 = pyo.SOSConstraint(var=m.x, sos=2)
        opt.add_sos_constraint(m.c2)
        self.assertEqual(opt._solver_model.getAttr('NumSOS'), 1)
        opt.remove_sos_constraint(m.c2)
        opt.update()
        self.assertEqual(opt._solver_model.getAttr('NumSOS'), 1)

    def test_update7(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()

        opt = pyo.SolverFactory('gurobi_persistent')
        opt.set_instance(m)
        self.assertEqual(opt._solver_model.getAttr('NumVars'), 0)

        opt.remove_var(m.x)
        opt.update()
        self.assertEqual(opt._solver_model.getAttr('NumVars'), 1)

        opt.add_var(m.x)
        self.assertEqual(opt._solver_model.getAttr('NumVars'), 1)
        opt.update()
        self.assertEqual(opt._solver_model.getAttr('NumVars'), 2)

        opt.remove_var(m.x)
        opt.update()
        opt.add_var(m.x)
        opt.remove_var(m.x)
        opt.update()
        self.assertEqual(opt._solver_model.getAttr('NumVars'), 1)

    def test_linear_constraint_attr(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.c = pyo.Constraint(expr=m.x + m.y == 1)

        opt = pyo.SolverFactory('gurobi_persistent')
        opt.set_instance(m)
        opt.set_linear_constraint_attr(m.c, 'Lazy', 1)
        self.assertEqual(opt.get_linear_constraint_attr(m.c, 'Lazy'), 1)

    def test_quadratic_constraint_attr(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.c = pyo.Constraint(expr=m.y >= m.x**2)

        opt = pyo.SolverFactory('gurobi_persistent')
        opt.set_instance(m)
        self.assertEqual(opt.get_quadratic_constraint_attr(m.c, 'QCRHS'), 0)

    def test_var_attr(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var(within=pyo.Binary)

        opt = pyo.SolverFactory('gurobi_persistent')
        opt.set_instance(m)
        opt.set_var_attr(m.x, 'Start', 1)
        self.assertEqual(opt.get_var_attr(m.x, 'Start'), 1)

    def test_callback(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var(bounds=(0, 4))
        m.y = pyo.Var(within=pyo.Integers, bounds=(0, None))
        m.obj = pyo.Objective(expr=2 * m.x + m.y)
        m.cons = pyo.ConstraintList()

        def _add_cut(xval):
            m.x.value = xval
            return m.cons.add(m.y >= taylor_series_expansion((m.x - 2) ** 2))

        _add_cut(0)
        _add_cut(4)

        opt = pyo.SolverFactory('gurobi_persistent')
        opt.set_instance(m)
        opt.set_gurobi_param('PreCrush', 1)
        opt.set_gurobi_param('LazyConstraints', 1)

        def _my_callback(cb_m, cb_opt, cb_where):
            if cb_where == gurobipy.GRB.Callback.MIPSOL:
                cb_opt.cbGetSolution(vars=[m.x, m.y])
                if m.y.value < (m.x.value - 2) ** 2 - 1e-6:
                    cb_opt.cbLazy(_add_cut(m.x.value))

        opt.set_callback(_my_callback)
        opt.solve()
        self.assertAlmostEqual(m.x.value, 1)
        self.assertAlmostEqual(m.y.value, 1)

    def test_add_column(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var(within=pyo.NonNegativeReals)
        m.c = pyo.Constraint(expr=(0, m.x, 1))
        m.obj = pyo.Objective(expr=-m.x)

        opt = pyo.SolverFactory('gurobi_persistent')
        opt.set_instance(m)
        opt.solve()
        self.assertAlmostEqual(m.x.value, 1)

        m.y = pyo.Var(within=pyo.NonNegativeReals)

        opt.add_column(m, m.y, -3, [m.c], [2])
        opt.solve()

        self.assertAlmostEqual(m.x.value, 0)
        self.assertAlmostEqual(m.y.value, 0.5)

    def test_add_column_exceptions(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.c = pyo.Constraint(expr=(0, m.x, 1))
        m.ci = pyo.Constraint([1, 2], rule=lambda m, i: (0, m.x, i + 1))
        m.cd = pyo.Constraint(expr=(0, -m.x, 1))
        m.cd.deactivate()
        m.obj = pyo.Objective(expr=-m.x)

        opt = pyo.SolverFactory('gurobi_persistent')

        # set_instance not called
        self.assertRaises(RuntimeError, opt.add_column, m, m.x, 0, [m.c], [1])

        opt.set_instance(m)

        m2 = pyo.ConcreteModel()
        m2.y = pyo.Var()
        m2.c = pyo.Constraint(expr=(0, m.x, 1))

        # different model than attached to opt
        self.assertRaises(RuntimeError, opt.add_column, m2, m2.y, 0, [], [])
        # pyomo var attached to different model
        self.assertRaises(RuntimeError, opt.add_column, m, m2.y, 0, [], [])

        z = pyo.Var()
        # pyomo var floating
        self.assertRaises(RuntimeError, opt.add_column, m, z, -2, [m.c, z], [1])

        m.y = pyo.Var()
        # len(coefficients) == len(constraints)
        self.assertRaises(RuntimeError, opt.add_column, m, m.y, -2, [m.c], [1, 2])
        self.assertRaises(RuntimeError, opt.add_column, m, m.y, -2, [m.c, z], [1])

        # add indexed constraint
        self.assertRaises(AttributeError, opt.add_column, m, m.y, -2, [m.ci], [1])
        # add something not a ConstraintData
        self.assertRaises(AttributeError, opt.add_column, m, m.y, -2, [m.x], [1])

        # constraint not on solver model
        self.assertRaises(KeyError, opt.add_column, m, m.y, -2, [m2.c], [1])

        # inactive constraint
        self.assertRaises(KeyError, opt.add_column, m, m.y, -2, [m.cd], [1])

        opt.add_var(m.y)
        # var already in solver model
        self.assertRaises(RuntimeError, opt.add_column, m, m.y, -2, [m.c], [1])



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.extensions import ExtensionBuilderFactory
from pyomo.opt import SolverFactory
from .build import PyNumeroBuilder
from .algorithms.solvers.cyipopt_solver import PyomoCyIpoptSolver
from .algorithms.solvers.scipy_solvers import (
    PyomoFsolveSolver,
    PyomoRootSolver,
    PyomoNewtonSolver,
    PyomoSecantNewtonSolver,
)


def load():
    ExtensionBuilderFactory.register('pynumero')(PyNumeroBuilder)
    SolverFactory.register(
        'cyipopt', doc='Cyipopt: direct python bindings to the Ipopt NLP solver'
    )(PyomoCyIpoptSolver)
    SolverFactory.register(
        "scipy.fsolve",
        doc=("fsolve: A SciPy wrapper around MINPACK's hybrd and hybrj algorithms"),
    )(PyomoFsolveSolver)
    SolverFactory.register(
        "scipy.root", doc=("root: Find the root of a vector function")
    )(PyomoRootSolver)
    SolverFactory.register(
        "scipy.newton", doc="newton: Find a zero of a scalar-valued function"
    )(PyomoNewtonSolver)
    SolverFactory.register(
        "scipy.secant-newton",
        doc=(
            "secant-newton: Take a few secant iterations to try to converge"
            " a potentially linear equation quickly, then switch to Newton's"
            " method"
        ),
    )(PyomoSecantNewtonSolver)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from collections import defaultdict


class HashDispatcher(defaultdict):
    """Dispatch table for generating "universal" hashing of all Python objects.

    This class manages a dispatch table for providing hash functions for all Python
    types.  When an object is passed to the Hasher, it determines the appropriate
    hashing strategy based on the object's type:

      - If a custom hashing function is registered for the type, it is used.
      - If the object is natively hashable, the default hash is used.
      - If the object is unhashable, the object's :func:`id()` is used as a fallback.

    The Hasher also includes special handling for tuples by recursively applying the
    appropriate hashing strategy to each element within the tuple.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(lambda: self._missing_impl, *args, **kwargs)
        self[tuple] = self._tuple

    def _missing_impl(self, val):
        try:
            hash(val)
            self[val.__class__] = self._hashable
        except:
            self[val.__class__] = self._unhashable
        return self[val.__class__](val)

    @staticmethod
    def _hashable(val):
        return val

    @staticmethod
    def _unhashable(val):
        return id(val)

    def _tuple(self, val):
        return tuple(self[i.__class__](i) for i in val)

    def hashable(self, obj, hashable=None):
        if isinstance(obj, type):
            cls = obj
        else:
            cls = type(obj)
        if hashable is None:
            fcn = self.get(cls, None)
            if fcn is None:
                raise KeyError(obj)
            return fcn is self._hashable
        self[cls] = self._hashable if hashable else self._unhashable


#: The global 'hasher' instance for managing "universal" hashing.
#:
#: This instance of the :class:`HashDispatcher` is used by
#: :class:`~pyomo.common.collections.component_map.ComponentMap` and
#: :class:`~pyomo.common.collections.component_set.ComponentSet` for
#: generating hashes for all Python and Pyomo types.
hasher = HashDispatcher()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""
Cutting plane-based GDP reformulation.

Implements a general cutting plane-based reformulation for linear and
convex GDPs.
"""

from pyomo.common.config import (
    ConfigBlock,
    ConfigValue,
    NonNegativeFloat,
    PositiveInt,
    In,
)
from pyomo.common.modeling import unique_component_name
from pyomo.core import (
    Any,
    Block,
    Constraint,
    Objective,
    Param,
    Var,
    SortComponents,
    Transformation,
    TransformationFactory,
    value,
    NonNegativeIntegers,
    Reals,
    NonNegativeReals,
    Suffix,
    ComponentMap,
)
from pyomo.core.expr import differentiate
from pyomo.common.collections import ComponentSet
from pyomo.opt import SolverFactory
from pyomo.repn import generate_standard_repn

from pyomo.gdp import Disjunct, Disjunction, GDP_Error
from pyomo.gdp.util import (
    verify_successful_solve,
    NORMAL,
    clone_without_expression_components,
)

from pyomo.contrib.fme.fourier_motzkin_elimination import (
    Fourier_Motzkin_Elimination_Transformation,
)

import logging

logger = logging.getLogger('pyomo.gdp.cuttingplane')


def do_not_tighten(m):
    return m


def _get_constraint_exprs(constraints, hull_to_bigm_map):
    """Returns a list of expressions which are constrain.expr translated
    into the bigm space, for each constraint in constraints.
    """
    cuts = []
    for cons in constraints:
        cuts.append(
            clone_without_expression_components(cons.expr, substitute=hull_to_bigm_map)
        )
    return cuts


def _constraint_tight(model, constraint, TOL):
    """
    Returns a list [a,b] where a is -1 if the lower bound is tight or
    slightly violated, b is 1 if the upper bound is tight of slightly
    violated, and [a,b]=[-1,1] if we have an exactly satisfied (or
    slightly violated) equality.
    """
    val = value(constraint.body)
    ans = [0, 0]
    if constraint.lower is not None:
        if val - value(constraint.lower) <= TOL:
            # tight or in violation of LB
            ans[0] -= 1

    if constraint.upper is not None:
        if value(constraint.upper) - val <= TOL:
            # tight or in violation of UB
            ans[1] += 1

    return ans


def _get_linear_approximation_expr(normal_vec, point):
    """Returns constraint linearly approximating constraint normal to normal_vec
    at point"""
    body = 0
    for coef, v in zip(point, normal_vec):
        body -= coef * v
    return body >= -sum(normal_vec[idx] * v.value for (idx, v) in enumerate(point))


def _precompute_potentially_useful_constraints(transBlock_rHull, disaggregated_vars):
    instance_rHull = transBlock_rHull.model()
    constraints = transBlock_rHull.constraints_for_FME = []
    for constraint in instance_rHull.component_data_objects(
        Constraint, active=True, descend_into=Block, sort=SortComponents.deterministic
    ):
        # we don't care about anything that does not involve at least one
        # disaggregated variable.
        repn = generate_standard_repn(constraint.body)
        for v in repn.linear_vars + repn.quadratic_vars + repn.nonlinear_vars:
            # ESJ: This is why disaggregated_vars is a ComponentSet
            if v in disaggregated_vars:
                constraints.append(constraint)
                break


def create_cuts_fme(
    transBlock_rHull,
    var_info,
    hull_to_bigm_map,
    rBigM_linear_constraints,
    rHull_vars,
    disaggregated_vars,
    norm,
    cut_threshold,
    zero_tolerance,
    integer_arithmetic,
    constraint_tolerance,
):
    """Returns a cut which removes x* from the relaxed bigm feasible region.

    Finds all the constraints which are tight at xhat (assumed to be the
    solution currently in instance_rHull), and calculates a composite normal
    vector by summing the vectors normal to each of these constraints. Then
    Fourier-Motzkin elimination is used to project the disaggregated variables
    out of the polyhedron formed by the composite normal and the collection
    of tight constraints. This results in multiple cuts, of which we select
    one that cuts of x* by the greatest margin, as long as that margin is
    more than cut_threshold. If no cut satisfies the margin specified by
    cut_threshold, we return None.

    Parameters
    -----------
    transBlock_rHull: transformation block on relaxed hull instance
    var_info: List of tuples (rBigM_var, rHull_var, xstar_param)
    hull_to_bigm_map: For expression substitution, maps id(hull_var) to
                      corresponding bigm var
    rBigM_linear_constraints: list of linear constraints in relaxed bigM
    rHull_vars: list of all variables in relaxed hull
    disaggregated_vars: ComponentSet of disaggregated variables in hull
                        reformulation
    norm: norm used in the separation problem
    cut_threshold: Amount x* needs to be infeasible in generated cut in order
                   to consider the cut for addition to the bigM model.
    zero_tolerance: Tolerance at which a float will be treated as 0 during
                    Fourier-Motzkin elimination
    integer_arithmetic: boolean, whether or not to require Fourier-Motzkin
                        Elimination does integer arithmetic. Only possible
                        when all data is integer.
    constraint_tolerance: Tolerance at which we will consider a constraint
                          tight.
    """
    instance_rHull = transBlock_rHull.model()
    # In the first iteration, we will compute a list of constraints that could
    # ever be interesting: Everything that involves at least one disaggregated
    # variable.
    if transBlock_rHull.component("constraints_for_FME") is None:
        _precompute_potentially_useful_constraints(transBlock_rHull, disaggregated_vars)

    tight_constraints = Block()
    conslist = tight_constraints.constraints = Constraint(NonNegativeIntegers)
    conslist.construct()
    something_interesting = False
    for constraint in transBlock_rHull.constraints_for_FME:
        multipliers = _constraint_tight(
            instance_rHull, constraint, constraint_tolerance
        )
        for multiplier in multipliers:
            if multiplier:
                something_interesting = True
                f = constraint.body
                firstDerivs = differentiate(f, wrt_list=rHull_vars)
                normal_vec = [multiplier * value(_) for _ in firstDerivs]
                # check if constraint is linear
                if f.polynomial_degree() == 1:
                    conslist[len(conslist)] = constraint.expr
                else:
                    # we will use the linear approximation of this constraint at
                    # x_hat
                    conslist[len(conslist)] = _get_linear_approximation_expr(
                        normal_vec, rHull_vars
                    )

    # NOTE: we now have all the tight Constraints (in the pyomo sense of the
    # word "Constraint"), but we are missing some variable bounds. The ones for
    # the disaggregated variables will be added by FME

    # It is possible that the separation problem returned a point in the
    # interior of the convex hull. It is also possible that the only active
    # constraints do not involve the disaggregated variables. In these
    # situations, there are not constraints from which to create a valid cut.
    if not something_interesting:
        return None

    tight_constraints.construct()
    logger.info(
        "Calling FME transformation on %s constraints to eliminate"
        " %s variables" % (len(tight_constraints.constraints), len(disaggregated_vars))
    )
    TransformationFactory('contrib.fourier_motzkin_elimination').apply_to(
        tight_constraints,
        vars_to_eliminate=disaggregated_vars,
        zero_tolerance=zero_tolerance,
        do_integer_arithmetic=integer_arithmetic,
        projected_constraints_name="fme_constraints",
    )
    fme_results = tight_constraints.fme_constraints
    projected_constraints = [cons for i, cons in fme_results.items()]

    # we created these constraints with the variables from rHull. We
    # actually need constraints for BigM and rBigM now!
    cuts = _get_constraint_exprs(projected_constraints, hull_to_bigm_map)

    # We likely have some cuts that duplicate other constraints now. We will
    # filter them to make sure that they do in fact cut off x*. If that's the
    # case, we know they are not already in the BigM relaxation. Because they
    # came from FME, they are very likely redundant, so we'll keep the best one
    # we find
    best = 0
    best_cut = None
    cuts_to_keep = []
    for i, cut in enumerate(cuts):
        # x* is still in rBigM, so we can just remove this constraint if it
        # is satisfied at x*
        logger.info("FME: Post-processing cut %s" % cut)
        if value(cut):
            logger.info("FME:\t Doesn't cut off x*")
            continue
        # we have found a constraint which cuts of x* by some convincing amount
        # and is not already in rBigM.
        cuts_to_keep.append(i)
        # We know cut is lb <= expr and that it's violated
        assert len(cut.args) == 2
        cut_off = value(cut.args[0]) - value(cut.args[1])
        if cut_off > cut_threshold and cut_off > best:
            best = cut_off
            best_cut = cut
            logger.info("FME:\t New best cut: Cuts off x* by %s." % best)

    # NOTE: this is not used right now, but it's not hard to imagine a world in
    # which we would want to keep multiple cuts from FME, so leaving it in for
    # now.
    cuts = [cuts[i] for i in cuts_to_keep]

    if best_cut is not None:
        return [best_cut]

    return None


def create_cuts_normal_vector(
    transBlock_rHull,
    var_info,
    hull_to_bigm_map,
    rBigM_linear_constraints,
    rHull_vars,
    disaggregated_vars,
    norm,
    cut_threshold,
    zero_tolerance,
    integer_arithmetic,
    constraint_tolerance,
):
    """Returns a cut which removes x* from the relaxed bigm feasible region.

    Ignores all parameters except var_info and cut_threshold, and constructs
    a cut at x_hat, the projection of the relaxed bigM solution x* onto the hull,
    which is perpendicular to the vector from x* to x_hat.

    Note that this method will often lead to numerical difficulties since both
    x* and x_hat are solutions to optimization problems. To mitigate this,
    use some method of backing off the cut to make it a bit more conservative.

    Parameters
    -----------
    transBlock_rHull: transformation block on relaxed hull instance. Ignored by
                      this callback.
    var_info: List of tuples (rBigM_var, rHull_var, xstar_param)
    hull_to_bigm_map: For expression substitution, maps id(hull_var) to
                      corresponding bigm var. Ignored by this callback
    rBigM_linear_constraints: list of linear constraints in relaxed bigM.
                              Ignored by this callback.
    rHull_vars: list of all variables in relaxed hull. Ignored by this callback.
    disaggregated_vars: ComponentSet of disaggregated variables in hull
                        reformulation. Ignored by this callback
    norm: The norm used in the separation problem, will be used to calculate
          the subgradient used to generate the cut
    cut_threshold: Amount x* needs to be infeasible in generated cut in order
                   to consider the cut for addition to the bigM model.
    zero_tolerance: Tolerance at which a float will be treated as 0 during
                    Fourier-Motzkin elimination. Ignored by this callback
    integer_arithmetic: Ignored by this callback (specifies FME use integer
                        arithmetic)
    constraint_tolerance: Ignored by this callback (specifies when constraints
                          are considered tight in FME)
    """
    cutexpr = 0
    if norm == 2:
        for x_rbigm, x_hull, x_star in var_info:
            cutexpr += (x_hull.value - x_star.value) * (x_rbigm - x_hull.value)
    elif norm == float('inf'):
        duals = transBlock_rHull.model().dual
        if len(duals) == 0:
            raise GDP_Error(
                "No dual information in the separation problem! "
                "To use the infinity norm and the "
                "create_cuts_normal_vector method, you must use "
                "a solver which provides dual information."
            )
        i = 0
        for x_rbigm, x_hull, x_star in var_info:
            # ESJ: We wrote this so duals will be nonnegative
            mu_plus = value(duals[transBlock_rHull.inf_norm_linearization[i]])
            mu_minus = value(duals[transBlock_rHull.inf_norm_linearization[i + 1]])
            assert mu_plus >= 0
            assert mu_minus >= 0
            cutexpr += (mu_plus - mu_minus) * (x_rbigm - x_hull.value)
            i += 2

    # make sure we're cutting off x* by enough.
    if value(cutexpr) < -cut_threshold:
        return [cutexpr >= 0]
    logger.warning(
        "Generated cut did not remove relaxed BigM solution by more "
        "than the specified threshold of %s. Stopping cut "
        "generation." % cut_threshold
    )
    return None


def back_off_constraint_with_calculated_cut_violation(
    cut, transBlock_rHull, bigm_to_hull_map, opt, stream_solver, TOL
):
    """Calculates the maximum violation of cut subject to the relaxed hull
    constraints. Increases this violation by TOL (to account for optimality
    tolerance in solving the problem), and, if it finds that cut can be violated
    up to this tolerance, makes it more conservative such that it no longer can.

    Parameters
    ----------
    cut: The cut to be made more conservative, a Constraint
    transBlock_rHull: the relaxed hull model's transformation Block
    bigm_to_hull_map: Dictionary mapping ids of bigM variables to the
                      corresponding variables on the relaxed hull instance
    opt: SolverFactory object for solving the maximum violation problem
    stream_solver: Whether or not to set tee=True while solving the maximum
                   violation problem.
    TOL: An absolute tolerance to be added to the calculated cut violation,
         to account for optimality tolerance in the maximum violation problem
         solve.
    """
    instance_rHull = transBlock_rHull.model()
    logger.info("Post-processing cut: %s" % cut.expr)
    # Take a constraint. We will solve a problem maximizing its violation
    # subject to rHull. We will add some user-specified tolerance to that
    # violation, and then add that much padding to it if it can be violated.
    transBlock_rHull.separation_objective.deactivate()

    transBlock_rHull.infeasibility_objective = Objective(
        expr=clone_without_expression_components(cut.body, substitute=bigm_to_hull_map)
    )

    results = opt.solve(instance_rHull, tee=stream_solver, load_solutions=False)
    if verify_successful_solve(results) is not NORMAL:
        logger.warning(
            "Problem to determine how much to "
            "back off the new cut "
            "did not solve normally. Leaving the constraint as is, "
            "which could lead to numerical trouble%s" % (results,)
        )
        # restore the objective
        transBlock_rHull.del_component(transBlock_rHull.infeasibility_objective)
        transBlock_rHull.separation_objective.activate()
        return
    instance_rHull.solutions.load_from(results)

    # we're minimizing, val is <= 0
    val = value(transBlock_rHull.infeasibility_objective) - TOL
    if val <= 0:
        logger.info("\tBacking off cut by %s" % val)
        lb, body, ub = cut.to_bounded_expression()
        cut.set_value((lb, body + abs(val), ub))
    # else there is nothing to do: restore the objective
    transBlock_rHull.del_component(transBlock_rHull.infeasibility_objective)
    transBlock_rHull.separation_objective.activate()


def back_off_constraint_by_fixed_tolerance(
    cut, transBlock_rHull, bigm_to_hull_map, opt, stream_solver, TOL
):
    """Makes cut more conservative by absolute tolerance TOL

    Parameters
    ----------
    cut: the cut to be made more conservative, a Constraint
    transBlock_rHull: the relaxed hull model's transformation Block. Ignored by
                      this callback
    bigm_to_hull_map: Dictionary mapping ids of bigM variables to the
                      corresponding variables on the relaxed hull instance.
                      Ignored by this callback.
    opt: SolverFactory object. Ignored by this callback
    stream_solver: Whether or not to set tee=True while solving. Ignored by
                   this callback
    TOL: An absolute tolerance to be added to make cut more conservative.
    """
    lb, body, ub = cut.to_bounded_expression()
    cut.set_value((lb, body + TOL, ub))


@TransformationFactory.register(
    'gdp.cuttingplane',
    doc="Relaxes a linear disjunctive model by "
    "adding cuts from convex hull to Big-M "
    "reformulation.",
)
class CuttingPlane_Transformation(Transformation):
    """Relax convex disjunctive model by forming the bigm relaxation and then
    iteratively adding cuts from the hull relaxation (or the hull relaxation
    after some basic steps) in order to strengthen the formulation.

    Note that gdp.cuttingplane is not a structural transformation: If variables
    on the model are fixed, they will be treated as data, and unfixing them
    after transformation will very likely result in an invalid model.

    This transformation accepts the following keyword arguments:

    Parameters
    ----------
    solver : Solver name (as string) to use to solve relaxed BigM and separation
             problems
    solver_options : dictionary of options to pass to the solver
    stream_solver : Whether or not to display solver output
    verbose : Enable verbose output from cuttingplanes algorithm
    cuts_name : Optional name for the IndexedConstraint containing the projected
                cuts (must be a unique name with respect to the instance)
    minimum_improvement_threshold : Stopping criterion based on improvement in
                                    Big-M relaxation. This is the minimum
                                    difference in relaxed BigM objective
                                    values between consecutive iterations
    separation_objective_threshold : Stopping criterion based on separation
                                     objective. If separation objective is not
                                     at least this large, cut generation will
                                     terminate.
    cut_filtering_threshold : Stopping criterion based on effectiveness of the
                              generated cut: This is the amount by which
                              a cut must be violated at the relaxed bigM
                              solution in order to be added to the bigM model
    max_number_of_cuts : The maximum number of cuts to add to the big-M model
    norm : norm to use in the objective of the separation problem
    tighten_relaxation : callback to modify the GDP model before the hull
                         relaxation is taken (e.g. could be used to perform
                         basic steps)
    create_cuts : callback to create cuts using the solved relaxed bigM and hull
                  problems
    post_process_cut : callback to perform post-processing on created cuts
    back_off_problem_tolerance : tolerance to use while post-processing
    zero_tolerance : Tolerance at which a float will be considered 0 when
                     using Fourier-Motzkin elimination to create cuts.
    do_integer_arithmetic : Whether or not to require Fourier-Motzkin elimination
                            to do integer arithmetic. Only possible when all
                            data is integer.
    tight_constraint_tolerance : Tolerance at which a constraint is considered
                                 tight for the Fourier-Motzkin cut generation
                                 procedure

    By default, the callbacks will be set such that the algorithm performed is
    as presented in [1], but with an additional post-processing procedure to
    reduce numerical error, which calculates the maximum violation of the cut
    subject to the relaxed hull constraints, and then pads the constraint by
    this violation plus an additional user-specified tolerance.

    In addition, the create_cuts_fme function provides an (exponential time)
    method of generating cuts which reduces numerical error (and can eliminate
    it if all data is integer). It collects the hull constraints which are
    tight at the solution of the separation problem. It creates a cut in the
    extended space perpendicular to  a composite normal vector created by
    summing the directions normal to these constraints. It then performs
    fourier-motzkin elimination on the collection of constraints and the cut
    to project out the disaggregated variables. The resulting constraint which
    is most violated by the relaxed bigM solution is then returned.

    References
    ----------
        [1] Sawaya, N. W., Grossmann, I. E. (2005). A cutting plane method for
        solving linear generalized disjunctive programming problems. Computers
        and Chemical Engineering, 29, 1891-1913
    """

    CONFIG = ConfigBlock("gdp.cuttingplane")
    CONFIG.declare(
        'solver',
        ConfigValue(
            default='ipopt',
            domain=str,
            description="""Solver to use for relaxed BigM problem and the separation
        problem""",
            doc="""
        This specifies the solver which will be used to solve LP relaxation
        of the BigM problem and the separation problem. Note that this solver
        must be able to handle a quadratic objective because of the separation
        problem.
        """,
        ),
    )
    CONFIG.declare(
        'minimum_improvement_threshold',
        ConfigValue(
            default=0.01,
            domain=NonNegativeFloat,
            description="Threshold value for difference in relaxed bigM problem "
            "objectives used to decide when to stop adding cuts",
            doc="""
        If the difference between the objectives in two consecutive iterations is
        less than this value, the algorithm terminates without adding the cut
        generated in the last iteration.  
        """,
        ),
    )
    CONFIG.declare(
        'separation_objective_threshold',
        ConfigValue(
            default=0.01,
            domain=NonNegativeFloat,
            description="Threshold value used to decide when to stop adding cuts: "
            "If separation problem objective is not at least this quantity, cut "
            "generation will terminate.",
            doc="""
        If the separation problem objective (distance between relaxed bigM 
        solution and its projection onto the relaxed hull feasible region)
        does not exceed this threshold, the algorithm will terminate.
        """,
        ),
    )
    CONFIG.declare(
        'max_number_of_cuts',
        ConfigValue(
            default=100,
            domain=PositiveInt,
            description="The maximum number of cuts to add before the algorithm "
            "terminates.",
            doc="""
        If the algorithm does not terminate due to another criterion first,
        cut generation will stop after adding this many cuts.
        """,
        ),
    )
    CONFIG.declare(
        'norm',
        ConfigValue(
            default=2,
            domain=In([2, float('inf')]),
            description="Norm to use in the separation problem: 2, or float('inf')",
            doc="""
        Norm used to calculate distance in the objective of the separation 
        problem which finds the nearest point on the hull relaxation region
        to the current solution of the relaxed bigm problem.

        Supported norms are the Euclidean norm (specify 2) and the infinity 
        norm (specify float('inf')). Note that the first makes the separation 
        problem objective quadratic and the latter makes it linear.
        """,
        ),
    )
    CONFIG.declare(
        'verbose',
        ConfigValue(
            default=False,
            domain=bool,
            description="Flag to enable verbose output",
            doc="""
        If True, prints subproblem solutions, as well as potential and added cuts
        during algorithm.

        If False, only the relaxed BigM objective and minimal information about 
        cuts is logged.
        """,
        ),
    )
    CONFIG.declare(
        'stream_solver',
        ConfigValue(
            default=False,
            domain=bool,
            description="""If true, sets tee=True for every solve performed over
        "the course of the algorithm""",
        ),
    )
    CONFIG.declare(
        'solver_options',
        ConfigBlock(
            implicit=True,
            description="Dictionary of solver options",
            doc="""
        Dictionary of solver options that will be set for the solver for both the
        relaxed BigM and separation problem solves.
        """,
        ),
    )
    CONFIG.declare(
        'tighten_relaxation',
        ConfigValue(
            default=do_not_tighten,
            description="Callback which takes the GDP formulation and returns a "
            "GDP formulation with a tighter hull relaxation",
            doc="""
        Function which accepts the GDP formulation of the problem and returns
        a GDP formulation which the transformation will then take the hull
        reformulation of.

        Most typically, this callback would be used to apply basic steps before
        taking the hull reformulation, but anything which tightens the GDP can 
        be performed here.
        """,
        ),
    )
    CONFIG.declare(
        'create_cuts',
        ConfigValue(
            default=create_cuts_normal_vector,
            description="Callback which generates a list of cuts, given the solved "
            "relaxed bigM and relaxed hull solutions. If no cuts can be "
            "generated, returns None",
            doc="""
        Callback to generate cuts to be added to the bigM problem based on 
        solutions to the relaxed bigM problem and the separation problem.

        Arguments
        ---------
        transBlock_rBigm: transformation block on relaxed bigM instance
        transBlock_rHull: transformation block on relaxed hull instance
        var_info: List of tuples (rBigM_var, rHull_var, xstar_param)
        hull_to_bigm_map: For expression substitution, maps id(hull_var) to 
                          corresponding bigm var
        rBigM_linear_constraints: list of linear constraints in relaxed bigM
        rHull_vars: list of all variables in relaxed hull
        disaggregated_vars: ComponentSet of disaggregated variables in hull 
                            reformulation
        cut_threshold: Amount x* needs to be infeasible in generated cut in order
                       to consider the cut for addition to the bigM model.
        zero_tolerance: Tolerance at which a float will be treated as 0

        Returns
        -------
        list of cuts to be added to bigM problem (and relaxed bigM problem),
        represented as expressions using variables from the bigM model
        """,
        ),
    )
    CONFIG.declare(
        'post_process_cut',
        ConfigValue(
            default=back_off_constraint_with_calculated_cut_violation,
            description="Callback which takes a generated cut and post processes "
            "it, presumably to back it off in the case of numerical error. Set to "
            "None if not post-processing is desired.",
            doc="""
        Callback to adjust a cut returned from create_cuts before adding it to
        the model, presumably to make it more conservative in case of numerical
        error.

        Arguments
        ---------
        cut: the cut to be made more conservative, a Constraint
        transBlock_rHull: the relaxed hull model's transformation Block.
        bigm_to_hull_map: Dictionary mapping ids of bigM variables to the 
                          corresponding variables on the relaxed hull instance.
        opt: SolverFactory object for subproblem solves in this procedure
        stream_solver: Whether or not to set tee=True while solving.
        TOL: A tolerance

        Returns
        -------
        None, modifies the cut in place
        """,
        ),
    )
    # back off problem tolerance (on top of the solver's (sometimes))
    CONFIG.declare(
        'back_off_problem_tolerance',
        ConfigValue(
            default=1e-8,
            domain=NonNegativeFloat,
            description="Tolerance to pass to the post_process_cut callback.",
            doc="""
        Tolerance passed to the post_process_cut callback.

        Depending on the callback, different values could make sense, but 
        something on the order of the solver's optimality or constraint 
        tolerances is appropriate.
        """,
        ),
    )
    CONFIG.declare(
        'cut_filtering_threshold',
        ConfigValue(
            default=0.001,
            domain=NonNegativeFloat,
            description="Tolerance used to decide if a cut removes x* from the "
            "relaxed BigM problem by enough to be added to the bigM problem.",
            doc="""
        Absolute tolerance used to decide whether to keep a cut. We require
        that, when evaluated at x* (the relaxed BigM optimal solution), the 
        cut be infeasible by at least this tolerance.
        """,
        ),
    )
    CONFIG.declare(
        'zero_tolerance',
        ConfigValue(
            default=1e-9,
            domain=NonNegativeFloat,
            description="Tolerance at which floats are assumed to be 0 while "
            "performing Fourier-Motzkin elimination",
            doc="""
        Only relevant when create_cuts=create_cuts_fme, this sets the 
        zero_tolerance option for the Fourier-Motzkin elimination transformation.
        """,
        ),
    )
    CONFIG.declare(
        'do_integer_arithmetic',
        ConfigValue(
            default=False,
            domain=bool,
            description="Only relevant if using Fourier-Motzkin Elimination (FME) "
            "and if all problem data is integer, requires FME transformation to "
            "perform integer arithmetic to eliminate numerical error.",
            doc="""
        Only relevant when create_cuts=create_cuts_fme and if all problem data 
        is integer, this sets the do_integer_arithmetic flag to true for the 
        FME transformation, meaning that the projection to the big-M space 
        can be done with exact precision.
        """,
        ),
    )
    CONFIG.declare(
        'cuts_name',
        ConfigValue(
            default=None,
            domain=str,
            description="Optional name for the IndexedConstraint containing the "
            "projected cuts. Must be a unique name with respect to the "
            "instance.",
            doc="""
        Optional name for the IndexedConstraint containing the projected 
        constraints. If not specified, the cuts will be stored on a 
        private block created by the transformation, so if you want access 
        to them after the transformation, use this argument.

        Must be a string which is a unique component name with respect to the 
        Block on which the transformation is called.
        """,
        ),
    )
    CONFIG.declare(
        'tight_constraint_tolerance',
        ConfigValue(
            default=1e-6,  # Gurobi constraint tolerance
            domain=NonNegativeFloat,
            description="Tolerance at which a constraint is considered tight for "
            "the Fourier-Motzkin cut generation procedure.",
            doc="""
        For a constraint a^Tx <= b, the Fourier-Motzkin cut generation procedure
        will consider the constraint tight (and add it to the set of constraints
        being projected) when a^Tx - b is less than this tolerance. 

        It is recommended to set this tolerance to the constraint tolerance of
        the solver being used.
        """,
        ),
    )

    def __init__(self):
        super(CuttingPlane_Transformation, self).__init__()

    def _apply_to(self, instance, bigM=None, **kwds):
        original_log_level = logger.level
        log_level = logger.getEffectiveLevel()
        try:
            self._config = self.CONFIG(kwds.pop('options', {}))
            self._config.set_value(kwds)

            if self._config.verbose and log_level > logging.INFO:
                logger.setLevel(logging.INFO)
                self.verbose = True
            elif log_level <= logging.INFO:
                self.verbose = True
            else:
                self.verbose = False

            instance_rBigM, cuts_obj, instance_rHull, var_info, transBlockName = (
                self._setup_subproblems(instance, bigM, self._config.tighten_relaxation)
            )

            self._generate_cuttingplanes(
                instance_rBigM, cuts_obj, instance_rHull, var_info, transBlockName
            )

            # restore integrality
            TransformationFactory('core.relax_integer_vars').apply_to(
                instance, undo=True
            )
        finally:
            del self._config
            del self.verbose
            # restore logging level
            logger.setLevel(original_log_level)

    def _setup_subproblems(self, instance, bigM, tighten_relaxation_callback):
        # create transformation block
        transBlockName, transBlock = self._add_transformation_block(instance)

        # We store a list of all vars so that we can efficiently
        # generate maps among the subproblems

        transBlock.all_vars = list(
            v
            for v in instance.component_data_objects(
                Var, descend_into=(Block, Disjunct), sort=SortComponents.deterministic
            )
            if not v.is_fixed()
        )

        # we'll store all the cuts we add together
        nm = self._config.cuts_name
        if nm is None:
            cuts_obj = transBlock.cuts = Constraint(NonNegativeIntegers)
        else:
            # check that this really is an available name
            if instance.component(nm) is not None:
                raise GDP_Error(
                    "cuts_name was specified as '%s', but this is "
                    "already a component on the instance! Please "
                    "specify a unique name." % nm
                )
            instance.add_component(nm, Constraint(NonNegativeIntegers))
            cuts_obj = instance.component(nm)

        # get bigM and hull relaxations
        bigMRelaxation = TransformationFactory('gdp.bigm')
        hullRelaxation = TransformationFactory('gdp.hull')
        relaxIntegrality = TransformationFactory('core.relax_integer_vars')

        #
        # Generate the Hull relaxation (used for the separation
        # problem to generate cutting planes)
        #
        tighter_instance = tighten_relaxation_callback(instance)
        instance_rHull = hullRelaxation.create_using(tighter_instance)
        relaxIntegrality.apply_to(instance_rHull, transform_deactivated_blocks=True)

        #
        # Reformulate the instance using the BigM relaxation (this will
        # be the final instance returned to the user)
        #
        bigMRelaxation.apply_to(instance, bigM=bigM)

        #
        # Generate the continuous relaxation of the BigM transformation. We'll
        # restore it at the end.
        #
        relaxIntegrality.apply_to(instance, transform_deactivated_blocks=True)

        #
        # Add the xstar parameter for the Hull problem
        #
        transBlock_rHull = instance_rHull.component(transBlockName)
        #
        # this will hold the solution to rbigm each time we solve it. We
        # add it to the transformation block so that we don't have to
        # worry about name conflicts.
        transBlock_rHull.xstar = Param(
            range(len(transBlock.all_vars)), mutable=True, default=0, within=Reals
        )

        # we will add a block that we will deactivate to use to store the
        # extended space cuts. We never need to solve these, but we need them to
        # be constructed for the sake of Fourier-Motzkin Elimination
        extendedSpaceCuts = transBlock_rHull.extendedSpaceCuts = Block()
        extendedSpaceCuts.deactivate()
        extendedSpaceCuts.cuts = Constraint(Any)

        #
        # Generate the mapping between the variables on all the
        # instances and the xstar parameter.
        #
        var_info = [
            (
                v,  # this is the bigM variable
                transBlock_rHull.all_vars[i],
                transBlock_rHull.xstar[i],
            )
            for i, v in enumerate(transBlock.all_vars)
        ]

        # NOTE: we wait to add the separation objective to the rHull problem
        # because it is best to do it in the first iteration, so that we can
        # skip stale variables.

        return (instance, cuts_obj, instance_rHull, var_info, transBlockName)

    # this is the map that I need to translate my projected cuts and add
    # them to bigM
    def _create_hull_to_bigm_substitution_map(self, var_info):
        return dict((id(var_info[i][1]), var_info[i][0]) for i in range(len(var_info)))

    # this map is needed to solve the back-off problem for post-processing
    def _create_bigm_to_hull_substitution_map(self, var_info):
        return dict((id(var_info[i][0]), var_info[i][1]) for i in range(len(var_info)))

    def _get_disaggregated_vars(self, hull):
        disaggregatedVars = ComponentSet()
        for disjunction in hull.component_data_objects(
            Disjunction, descend_into=(Disjunct, Block)
        ):
            for disjunct in disjunction.disjuncts:
                transBlock = disjunct.transformation_block
                if transBlock is not None:
                    for v in transBlock.disaggregatedVars.component_data_objects(Var):
                        disaggregatedVars.add(v)

        return disaggregatedVars

    def _get_rBigM_obj_and_constraints(self, instance_rBigM):
        # We try to grab the first active objective. If there is more
        # than one, the writer will yell when we try to solve below. If
        # there are 0, we will yell here.
        rBigM_obj = next(
            instance_rBigM.component_data_objects(Objective, active=True), None
        )
        if rBigM_obj is None:
            raise GDP_Error(
                "Cannot apply cutting planes transformation "
                "without an active objective in the model!"
            )

        #
        # Collect all of the linear constraints that are in the rBigM
        # instance. We will need these so that we can compare what we get from
        # FME to them and make sure we aren't adding redundant constraints to
        # the model. For convenience, we will make sure they are all in the form
        # lb <= expr (so we will break equality constraints)
        #
        fme = TransformationFactory('contrib.fourier_motzkin_elimination')
        rBigM_linear_constraints = []
        for cons in instance_rBigM.component_data_objects(
            Constraint,
            descend_into=Block,
            sort=SortComponents.deterministic,
            active=True,
        ):
            body = cons.body
            if body.polynomial_degree() != 1:
                # We will never get a nonlinear constraint out of FME, so we
                # don't risk it being identical to this one.
                continue

            # TODO: Guess this shouldn't have been private...
            rBigM_linear_constraints.extend(fme._process_constraint(cons))

        # [ESJ Aug 13 2020] NOTE: We actually don't need to worry about variable
        # bounds here because the FME transformation will take care of them
        # (i.e. convert those of the disaggregated variables to constraints for
        # the purposes of the projection.)

        return rBigM_obj, rBigM_linear_constraints

    def _generate_cuttingplanes(
        self, instance_rBigM, cuts_obj, instance_rHull, var_info, transBlockName
    ):
        opt = SolverFactory(self._config.solver)
        stream_solver = self._config.stream_solver
        opt.options = dict(self._config.solver_options)

        improving = True
        prev_obj = None
        epsilon = self._config.minimum_improvement_threshold
        cuts = None

        transBlock_rHull = instance_rHull.component(transBlockName)

        rBigM_obj, rBigM_linear_constraints = self._get_rBigM_obj_and_constraints(
            instance_rBigM
        )

        # Get list of all variables in the rHull model which we will use when
        # calculating the composite normal vector.
        rHull_vars = [
            i
            for i in instance_rHull.component_data_objects(
                Var, descend_into=Block, sort=SortComponents.deterministic
            )
        ]

        # collect a list of disaggregated variables.
        disaggregated_vars = self._get_disaggregated_vars(instance_rHull)

        hull_to_bigm_map = self._create_hull_to_bigm_substitution_map(var_info)
        bigm_to_hull_map = self._create_bigm_to_hull_substitution_map(var_info)
        xhat = ComponentMap()

        while improving:
            # solve rBigM, solution is xstar
            results = opt.solve(instance_rBigM, tee=stream_solver, load_solutions=False)
            if verify_successful_solve(results) is not NORMAL:
                logger.warning(
                    "Relaxed BigM subproblem "
                    "did not solve normally. Stopping cutting "
                    "plane generation.\n\n%s" % (results,)
                )
                return
            instance_rBigM.solutions.load_from(results)

            rBigM_objVal = value(rBigM_obj)
            logger.warning("rBigM objective = %s" % (rBigM_objVal,))

            #
            # Add the separation objective to the hull subproblem if it's not
            # there already (so in the first iteration). We're waiting until now
            # to avoid it including variables that came back stale from the
            # rbigm solve.
            #
            if transBlock_rHull.component("separation_objective") is None:
                self._add_separation_objective(var_info, transBlock_rHull)

            # copy over xstar
            logger.info("x* is:")
            for x_rbigm, x_hull, x_star in var_info:
                if not x_rbigm.stale:
                    x_star.set_value(x_rbigm.value)
                    # initialize the X values
                    x_hull.set_value(x_rbigm.value, skip_validation=True)
                if self.verbose:
                    logger.info(
                        "\t%s = %s"
                        % (x_rbigm.getname(fully_qualified=True), x_rbigm.value)
                    )

            # compare objectives: check absolute difference close to 0, relative
            # difference further from 0.
            if prev_obj is None:
                improving = True
            else:
                obj_diff = prev_obj - rBigM_objVal
                improving = (
                    abs(obj_diff) > epsilon
                    if abs(rBigM_objVal) < 1
                    else abs(obj_diff / prev_obj) > epsilon
                )

            # solve separation problem to get xhat.
            results = opt.solve(instance_rHull, tee=stream_solver, load_solutions=False)
            if verify_successful_solve(results) is not NORMAL:
                logger.warning(
                    "Hull separation subproblem "
                    "did not solve normally. Stopping cutting "
                    "plane generation.\n\n%s" % (results,)
                )
                return
            instance_rHull.solutions.load_from(results)
            logger.warning(
                "separation problem objective value: %s"
                % value(transBlock_rHull.separation_objective)
            )

            # save xhat to initialize rBigM with in the next iteration
            if self.verbose:
                logger.info("xhat is: ")
            for x_rbigm, x_hull, x_star in var_info:
                xhat[x_rbigm] = value(x_hull)
                if self.verbose:
                    logger.info(
                        "\t%s = %s"
                        % (x_hull.getname(fully_qualified=True), x_hull.value)
                    )

            # [JDS 19 Dec 18] Note: we check that the separation objective was
            # significantly nonzero.  If it is too close to zero, either the
            # rBigM solution was in the convex hull, or the separation vector is
            # so close to zero that the resulting cut is likely to have
            # numerical issues.
            if (
                value(transBlock_rHull.separation_objective)
                < self._config.separation_objective_threshold
            ):
                logger.warning(
                    "Separation problem objective below threshold of"
                    " %s: Stopping cut generation."
                    % self._config.separation_objective_threshold
                )
                break

            cuts = self._config.create_cuts(
                transBlock_rHull,
                var_info,
                hull_to_bigm_map,
                rBigM_linear_constraints,
                rHull_vars,
                disaggregated_vars,
                self._config.norm,
                self._config.cut_filtering_threshold,
                self._config.zero_tolerance,
                self._config.do_integer_arithmetic,
                self._config.tight_constraint_tolerance,
            )

            # We are done if the cut generator couldn't return a valid cut
            if cuts is None:
                logger.warning("Did not generate a valid cut, stopping cut generation.")
                break
            if not improving:
                logger.warning(
                    "Difference in relaxed BigM problem objective "
                    "values from past two iterations is below "
                    "threshold of %s: Stopping cut generation." % epsilon
                )
                break

            for cut in cuts:
                # we add the cut to the model and then post-process it in place.
                cut_number = len(cuts_obj)
                logger.warning("Adding cut %s to BigM model." % (cut_number,))
                cuts_obj.add(cut_number, cut)
                if self._config.post_process_cut is not None:
                    self._config.post_process_cut(
                        cuts_obj[cut_number],
                        transBlock_rHull,
                        bigm_to_hull_map,
                        opt,
                        stream_solver,
                        self._config.back_off_problem_tolerance,
                    )

            if cut_number + 1 == self._config.max_number_of_cuts:
                logger.warning("Reached maximum number of cuts.")
                break

            prev_obj = rBigM_objVal

            # Initialize rbigm with xhat (for the next iteration)
            for x_rbigm, x_hull, x_star in var_info:
                x_rbigm.set_value(xhat[x_rbigm], skip_validation=True)

    def _add_transformation_block(self, instance):
        # creates transformation block with a unique name based on name, adds it
        # to instance, and returns it.
        transBlockName = unique_component_name(
            instance, '_pyomo_gdp_cuttingplane_transformation'
        )
        transBlock = Block()
        instance.add_component(transBlockName, transBlock)
        return transBlockName, transBlock

    def _add_separation_objective(self, var_info, transBlock_rHull):
        # creates the separation objective. That is just adding an objective for
        # Euclidean norm, it means adding an auxiliary variable to linearize the
        # L-infinity norm. We do this assuming that rBigM has been solved, and
        # if any variables come back stale, we leave them out of the separation
        # problem, as they aren't doing anything and they could cause numerical
        # issues later.

        # Deactivate any/all other objectives
        for o in transBlock_rHull.model().component_data_objects(Objective):
            o.deactivate()
        norm = self._config.norm
        to_delete = []

        if norm == 2:
            obj_expr = 0
            for i, (x_rbigm, x_hull, x_star) in enumerate(var_info):
                if not x_rbigm.stale:
                    obj_expr += (x_hull - x_star) ** 2
                else:
                    if self.verbose:
                        logger.info(
                            "The variable %s will not be included in "
                            "the separation problem: It was stale in "
                            "the rBigM solve." % x_rbigm.getname(fully_qualified=True)
                        )
                    to_delete.append(i)
        elif norm == float('inf'):
            u = transBlock_rHull.u = Var(domain=NonNegativeReals)
            inf_cons = transBlock_rHull.inf_norm_linearization = Constraint(
                NonNegativeIntegers
            )
            i = 0
            for j, (x_rbigm, x_hull, x_star) in enumerate(var_info):
                if not x_rbigm.stale:
                    # NOTE: these are written as >= constraints so that we know
                    # the duals will come back nonnegative.
                    inf_cons[i] = u - x_hull >= -x_star
                    inf_cons[i + 1] = u + x_hull >= x_star
                    i += 2
                else:
                    if self.verbose:
                        logger.info(
                            "The variable %s will not be included in "
                            "the separation problem: It was stale in "
                            "the rBigM solve." % x_rbigm.getname(fully_qualified=True)
                        )
                    to_delete.append(j)
            # we'll need the duals of these to get the subgradient
            self._add_dual_suffix(transBlock_rHull.model())
            obj_expr = u

        # delete the unneeded x_stars so that we don't add cuts involving
        # useless variables later.
        for i in sorted(to_delete, reverse=True):
            del var_info[i]

        # add separation objective to transformation block
        transBlock_rHull.separation_objective = Objective(expr=obj_expr)

    def _add_dual_suffix(self, rHull):
        # rHull is our model and we aren't giving it back (unless in the future
        # we we add a callback to do basic steps to it...), so we just check if
        # dual is there. If it's a Suffix, we'll borrow it. If it's something
        # else we'll rename it and add the Suffix.
        dual = rHull.component("dual")
        if dual is None:
            rHull.dual = Suffix(direction=Suffix.IMPORT)
        else:
            if dual.ctype is Suffix:
                return
            rHull.del_component(dual)
            rHull.dual = Suffix(direction=Suffix.IMPORT)
            rHull.add_component(unique_component_name(rHull, "dual"), dual)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import os
import pyomo.common.unittest as unittest
import pyomo.environ as pyo
import math

from pyomo.contrib.pynumero.dependencies import (
    numpy as np,
    numpy_available,
    scipy,
    scipy_available,
)
from pyomo.common.dependencies.scipy import sparse as spa

if not (numpy_available and scipy_available):
    raise unittest.SkipTest("Pynumero needs scipy and numpy to run NLP tests")

from pyomo.contrib.pynumero.asl import AmplInterface

if not AmplInterface.available():
    raise unittest.SkipTest(
        "Pynumero needs the ASL extension to run CyIpoptSolver tests"
    )

from pyomo.contrib.pynumero.algorithms.solvers.cyipopt_solver import CyIpoptSolver
from pyomo.contrib.pynumero.interfaces.cyipopt_interface import (
    CyIpoptNLP,
    cyipopt_available,
)

from pyomo.contrib.pynumero.interfaces.external_grey_box import (
    ExternalGreyBoxModel,
    ExternalGreyBoxBlock,
)
from pyomo.contrib.pynumero.interfaces.pyomo_nlp import PyomoGreyBoxNLP, PyomoNLP
from pyomo.contrib.pynumero.interfaces.pyomo_grey_box_nlp import (
    PyomoNLPWithGreyBoxBlocks,
)

from pyomo.contrib.pynumero.interfaces.tests.compare_utils import (
    check_vectors_specific_order,
    check_sparse_matrix_specific_order,
)


def create_pyomo_model(A1, A2, c1, c2, N, dt):
    m = pyo.ConcreteModel()

    # timesteps
    m.T = pyo.Set(initialize=list(range(N)), ordered=True)
    m.Tu = pyo.Set(initialize=list(range(N))[1:], ordered=True)

    # inputs (controls)
    m.F1 = pyo.Var(m.Tu, bounds=(0, None), initialize={t: 1 + 0.1 * t for t in m.Tu})
    m.F2 = pyo.Var(m.Tu, bounds=(0, None), initialize={t: 2 + 0.1 * t for t in m.Tu})

    # state variables
    m.h1 = pyo.Var(m.T, bounds=(0, None), initialize={t: 3 + 0.1 * t for t in m.T})
    m.h2 = pyo.Var(m.T, bounds=(0, None), initialize={t: 4 + 0.1 * t for t in m.T})

    # algebraics (outputs)
    m.F12 = pyo.Var(m.T, bounds=(0, None), initialize={t: 5 + 0.1 * t for t in m.T})
    m.Fo = pyo.Var(m.T, bounds=(0, None), initialize={t: 6 + 0.1 * t for t in m.T})

    @m.Constraint(m.Tu)
    def h1bal(m, t):
        return (m.h1[t] - m.h1[t - 1]) - dt / A1 * (
            m.F1[t] - c1 * pyo.sqrt(m.h1[t])
        ) == 0

    @m.Constraint(m.Tu)
    def h2bal(m, t):
        return (m.h2[t] - m.h2[t - 1]) - dt / A2 * (
            c1 * pyo.sqrt(m.h1[t]) + m.F2[t] - c2 * pyo.sqrt(m.h2[t])
        ) == 0

    @m.Constraint(m.T)
    def F12con(m, t):
        return c1 * pyo.sqrt(m.h1[t]) - m.F12[t] == 0

    @m.Constraint(m.T)
    def Focon(m, t):
        return c2 * pyo.sqrt(m.h2[t]) - m.Fo[t] == 0

    @m.Constraint(m.Tu)
    def min_inflow(m, t):
        return 2 <= m.F1[t]

    @m.Constraint(m.T)
    def max_outflow(m, t):
        return m.Fo[t] <= 4.5

    m.h10 = pyo.Constraint(expr=m.h1[m.T.first()] == 1.5)
    m.h20 = pyo.Constraint(expr=m.h2[m.T.first()] == 0.5)
    m.obj = pyo.Objective(
        expr=sum((m.h1[t] - 1.0) ** 2 + (m.h2[t] - 1.5) ** 2 for t in m.T)
    )

    return m


class TwoTanksSeries(ExternalGreyBoxModel):
    def __init__(self, A1, A2, c1, c2, N, dt):
        self._A1 = A1
        self._A2 = A2
        self._c1 = c1
        self._c2 = c2
        self._N = N
        self._dt = dt
        self._input_names = ['F1_{}'.format(t) for t in range(1, N)]
        self._input_names.extend(['F2_{}'.format(t) for t in range(1, N)])
        self._input_names.extend(['h1_{}'.format(t) for t in range(0, N)])
        self._input_names.extend(['h2_{}'.format(t) for t in range(0, N)])
        self._output_names = ['F12_{}'.format(t) for t in range(0, N)]
        self._output_names.extend(['Fo_{}'.format(t) for t in range(0, N)])
        self._equality_constraint_names = ['h1bal_{}'.format(t) for t in range(1, N)]
        self._equality_constraint_names.extend(
            ['h2bal_{}'.format(t) for t in range(1, N)]
        )

        # inputs
        self._F1 = np.zeros(N)  # we don't use the first one
        self._F2 = np.zeros(N)  # we don't use the first one
        self._h1 = np.zeros(N)
        self._h2 = np.zeros(N)

        # outputs
        self._F12 = np.zeros(N)
        self._Fo = np.zeros(N)

        # multipliers
        self._eq_con_mult_values = np.ones(2 * (N - 1))
        self._output_con_mult_values = np.ones(2 * N)

    def input_names(self):
        return self._input_names

    def equality_constraint_names(self):
        return self._equality_constraint_names

    def output_names(self):
        return self._output_names

    def finalize_block_construction(self, pyomo_block):
        # initialize all variables to 1 and set bounds
        for k in pyomo_block.inputs:
            pyomo_block.inputs[k].setlb(0)
            pyomo_block.inputs[k].value = 1.0
        for k in pyomo_block.outputs:
            pyomo_block.outputs[k].setlb(0)
            pyomo_block.outputs[k].value = 1.0

    def set_input_values(self, input_values):
        N = self._N
        assert len(input_values) == 4 * N - 2
        self._F1[1 : self._N] = np.copy(input_values[: N - 1])
        self._F2[1 : self._N] = np.copy(input_values[N - 1 : 2 * N - 2])
        self._h1 = np.copy(input_values[2 * N - 2 : 3 * N - 2])
        self._h2 = np.copy(input_values[3 * N - 2 : 4 * N - 2])

    def set_equality_constraint_multipliers(self, eq_con_multiplier_values):
        assert len(eq_con_multiplier_values) == 2 * (self._N - 1)
        np.copyto(self._eq_con_mult_values, eq_con_multiplier_values)

    def set_output_constraint_multipliers(self, output_con_multiplier_values):
        assert len(output_con_multiplier_values) == 2 * self._N
        np.copyto(self._output_con_mult_values, output_con_multiplier_values)

    def evaluate_equality_constraints(self):
        N = self._N
        F1 = self._F1
        F2 = self._F2
        h1 = self._h1
        h2 = self._h2

        resid = np.zeros(2 * (N - 1))

        for t in range(1, N):
            resid[t - 1] = (h1[t] - h1[t - 1]) - self._dt / self._A1 * (
                F1[t] - self._c1 * math.sqrt(h1[t])
            )

        for t in range(1, N):
            resid[t - 2 + N] = (h2[t] - h2[t - 1]) - self._dt / self._A2 * (
                self._c1 * math.sqrt(h1[t]) + F2[t] - self._c2 * math.sqrt(h2[t])
            )
        return resid

    def evaluate_outputs(self):
        N = self._N
        h1 = self._h1
        h2 = self._h2

        resid = np.zeros(2 * N)

        for t in range(N):
            resid[t] = self._c1 * math.sqrt(h1[t])

        for t in range(N):
            resid[t + N] = self._c2 * math.sqrt(h2[t])

        return resid

    def evaluate_jacobian_equality_constraints(self):
        N = self._N
        F1 = self._F1
        F2 = self._F2
        h1 = self._h1
        h2 = self._h2
        A1 = self._A1
        A2 = self._A2
        c1 = self._c1
        c2 = self._c2
        dt = self._dt

        nnz = 3 * (N - 1) + 4 * (N - 1)
        irow = np.zeros(nnz, dtype=np.int64)
        jcol = np.zeros(nnz, dtype=np.int64)
        data = np.zeros(nnz, dtype=np.float64)
        idx = 0
        # Jac h1bal
        for i in range(N - 1):
            irow[idx] = i
            jcol[idx] = i
            data[idx] = -dt / A1
            idx += 1
            irow[idx] = i
            jcol[idx] = 2 * (N - 1) + i
            data[idx] = -1
            idx += 1
            irow[idx] = i
            jcol[idx] = 2 * (N - 1) + i + 1
            data[idx] = 1 + dt / A1 * c1 * 1 / 2 * (h1[i + 1]) ** (-0.5)
            idx += 1
        # Jac h2bal
        for i in range(N - 1):
            irow[idx] = i + (N - 1)
            jcol[idx] = i + (N - 1)
            data[idx] = -dt / A2
            idx += 1
            irow[idx] = i + (N - 1)
            jcol[idx] = 2 * (N - 1) + i + 1
            data[idx] = -dt / A2 * c1 * 1 / 2 * (h1[i + 1]) ** (-0.5)
            idx += 1
            irow[idx] = i + (N - 1)
            jcol[idx] = 2 * (N - 1) + N + i
            data[idx] = -1
            idx += 1
            irow[idx] = i + (N - 1)
            jcol[idx] = 2 * (N - 1) + N + i + 1
            data[idx] = 1 + dt / A2 * c2 * 1 / 2 * (h2[i + 1]) ** (-0.5)
            idx += 1

        assert idx == nnz
        return spa.coo_matrix(
            (data, (irow, jcol)), shape=(2 * (N - 1), 2 * (N - 1) + 2 * N)
        )

    def evaluate_jacobian_outputs(self):
        N = self._N
        F1 = self._F1
        F2 = self._F2
        h1 = self._h1
        h2 = self._h2
        A1 = self._A1
        A2 = self._A2
        c1 = self._c1
        c2 = self._c2
        dt = self._dt

        nnz = 2 * N
        irow = np.zeros(nnz, dtype=np.int64)
        jcol = np.zeros(nnz, dtype=np.int64)
        data = np.zeros(nnz, dtype=np.float64)
        idx = 0
        # Jac F12
        for i in range(N):
            irow[idx] = i
            jcol[idx] = 2 * (N - 1) + i
            data[idx] = 1 / 2 * c1 * h1[i] ** (-0.5)
            idx += 1
        for i in range(N):
            irow[idx] = N + i
            jcol[idx] = 2 * (N - 1) + N + i
            data[idx] = 1 / 2 * c2 * h2[i] ** (-0.5)
            idx += 1

        assert idx == nnz
        return spa.coo_matrix((data, (irow, jcol)), shape=(2 * N, 2 * (N - 1) + 2 * N))

    def evaluate_hessian_equality_constraints(self):
        N = self._N
        F1 = self._F1
        F2 = self._F2
        h1 = self._h1
        h2 = self._h2
        A1 = self._A1
        A2 = self._A2
        c1 = self._c1
        c2 = self._c2
        dt = self._dt
        lam = self._eq_con_mult_values

        nnz = 2 * (N - 1)
        irow = np.zeros(nnz, dtype=np.int64)
        jcol = np.zeros(nnz, dtype=np.int64)
        data = np.zeros(nnz, dtype=np.float64)
        idx = 0
        for i in range(N - 1):
            irow[idx] = 2 * (N - 1) + i + 1
            jcol[idx] = 2 * (N - 1) + i + 1
            data[idx] = lam[i] * dt / A1 * (-c1 / 4) * h1[i + 1] ** (-1.5) + lam[
                (N - 1) + i
            ] * dt / A2 * (c1 / 4) * h1[i + 1] ** (-1.5)
            idx += 1
            irow[idx] = 2 * (N - 1) + N + i + 1
            jcol[idx] = 2 * (N - 1) + N + i + 1
            data[idx] = lam[(N - 1) + i] * dt / A2 * (-c2 / 4) * h2[i + 1] ** (-1.5)
            idx += 1

        assert idx == nnz
        hess = spa.coo_matrix(
            (data, (irow, jcol)), shape=(2 * (N - 1) + 2 * N, 2 * (N - 1) + 2 * N)
        )
        return hess

    def evaluate_hessian_outputs(self):
        N = self._N
        F1 = self._F1
        F2 = self._F2
        h1 = self._h1
        h2 = self._h2
        A1 = self._A1
        A2 = self._A2
        c1 = self._c1
        c2 = self._c2
        dt = self._dt
        lam = self._output_con_mult_values

        nnz = 2 * N
        irow = np.zeros(nnz, dtype=np.int64)
        jcol = np.zeros(nnz, dtype=np.int64)
        data = np.zeros(nnz, dtype=np.float64)
        idx = 0

        # Hess F12_t
        for i in range(N):
            irow[idx] = 2 * (N - 1) + i
            jcol[idx] = 2 * (N - 1) + i
            data[idx] = lam[i] * c1 * (-1 / 4) * h1[i] ** (-1.5)
            idx += 1
        # Hess Fo_t
        for i in range(N):
            irow[idx] = 2 * (N - 1) + N + i
            jcol[idx] = 2 * (N - 1) + N + i
            data[idx] = lam[N + i] * c2 * (-1 / 4) * h2[i] ** (-1.5)
            idx += 1

        assert idx == nnz
        hess = spa.coo_matrix(
            (data, (irow, jcol)), shape=(2 * (N - 1) + 2 * N, 2 * (N - 1) + 2 * N)
        )
        return hess


def create_pyomo_external_grey_box_model(A1, A2, c1, c2, N, dt):
    m2 = pyo.ConcreteModel()
    m2.T = pyo.Set(initialize=list(range(N)), ordered=True)
    m2.Tu = pyo.Set(initialize=list(range(N))[1:], ordered=True)
    m2.egb = ExternalGreyBoxBlock()
    m2.egb.set_external_model(TwoTanksSeries(A1, A2, c1, c2, N, dt))

    # initialize the same as the pyomo model
    for t in m2.Tu:
        m2.egb.inputs['F1_{}'.format(t)].value = 1 + 0.1 * t
        m2.egb.inputs['F2_{}'.format(t)].value = 2 + 0.1 * t
    for t in m2.T:
        m2.egb.inputs['h1_{}'.format(t)].value = 3 + 0.1 * t
        m2.egb.inputs['h2_{}'.format(t)].value = 4 + 0.1 * t
        m2.egb.outputs['F12_{}'.format(t)].value = 5 + 0.1 * t
        m2.egb.outputs['Fo_{}'.format(t)].value = 6 + 0.1 * t

    @m2.Constraint(m2.Tu)
    def min_inflow(m, t):
        F1_t = m.egb.inputs['F1_{}'.format(t)]
        return 2 <= F1_t

    @m2.Constraint(m2.T)
    def max_outflow(m, t):
        Fo_t = m.egb.outputs['Fo_{}'.format(t)]
        return Fo_t <= 4.5

    m2.h10 = pyo.Constraint(expr=m2.egb.inputs['h1_0'] == 1.5)
    m2.h20 = pyo.Constraint(expr=m2.egb.inputs['h2_0'] == 0.5)
    m2.obj = pyo.Objective(
        expr=sum(
            (m2.egb.inputs['h1_{}'.format(t)] - 1.0) ** 2
            + (m2.egb.inputs['h2_{}'.format(t)] - 1.5) ** 2
            for t in m2.T
        )
    )

    return m2


class TestGreyBoxModel(unittest.TestCase):
    @unittest.skipIf(
        not pyo.SolverFactory('ipopt').available(exception_flag=False),
        "Ipopt needed to run tests with solve",
    )
    def test_compare_evaluations(self):
        A1 = 5
        A2 = 10
        c1 = 3
        c2 = 4
        N = 6
        dt = 1

        m = create_pyomo_model(A1, A2, c1, c2, N, dt)
        solver = pyo.SolverFactory('ipopt')
        solver.options['linear_solver'] = 'mumps'
        status = solver.solve(m, tee=False)
        m_nlp = PyomoNLP(m)

        mex = create_pyomo_external_grey_box_model(A1, A2, c1, c2, N, dt)
        # mex_nlp = PyomoGreyBoxNLP(mex)
        mex_nlp = PyomoNLPWithGreyBoxBlocks(mex)

        # get the variable and constraint order and create the maps
        # reliable order independent comparisons
        m_x_order = m_nlp.primals_names()
        m_c_order = m_nlp.constraint_names()
        mex_x_order = mex_nlp.primals_names()
        mex_c_order = mex_nlp.constraint_names()

        x1list = [
            'h1[0]',
            'h1[1]',
            'h1[2]',
            'h1[3]',
            'h1[4]',
            'h1[5]',
            'h2[0]',
            'h2[1]',
            'h2[2]',
            'h2[3]',
            'h2[4]',
            'h2[5]',
            'F1[1]',
            'F1[2]',
            'F1[3]',
            'F1[4]',
            'F1[5]',
            'F2[1]',
            'F2[2]',
            'F2[3]',
            'F2[4]',
            'F2[5]',
            'F12[0]',
            'F12[1]',
            'F12[2]',
            'F12[3]',
            'F12[4]',
            'F12[5]',
            'Fo[0]',
            'Fo[1]',
            'Fo[2]',
            'Fo[3]',
            'Fo[4]',
            'Fo[5]',
        ]
        x2list = [
            'egb.inputs[h1_0]',
            'egb.inputs[h1_1]',
            'egb.inputs[h1_2]',
            'egb.inputs[h1_3]',
            'egb.inputs[h1_4]',
            'egb.inputs[h1_5]',
            'egb.inputs[h2_0]',
            'egb.inputs[h2_1]',
            'egb.inputs[h2_2]',
            'egb.inputs[h2_3]',
            'egb.inputs[h2_4]',
            'egb.inputs[h2_5]',
            'egb.inputs[F1_1]',
            'egb.inputs[F1_2]',
            'egb.inputs[F1_3]',
            'egb.inputs[F1_4]',
            'egb.inputs[F1_5]',
            'egb.inputs[F2_1]',
            'egb.inputs[F2_2]',
            'egb.inputs[F2_3]',
            'egb.inputs[F2_4]',
            'egb.inputs[F2_5]',
            'egb.outputs[F12_0]',
            'egb.outputs[F12_1]',
            'egb.outputs[F12_2]',
            'egb.outputs[F12_3]',
            'egb.outputs[F12_4]',
            'egb.outputs[F12_5]',
            'egb.outputs[Fo_0]',
            'egb.outputs[Fo_1]',
            'egb.outputs[Fo_2]',
            'egb.outputs[Fo_3]',
            'egb.outputs[Fo_4]',
            'egb.outputs[Fo_5]',
        ]
        x1_x2_map = dict(zip(x1list, x2list))
        x1idx_x2idx_map = {
            i: mex_x_order.index(x1_x2_map[m_x_order[i]]) for i in range(len(m_x_order))
        }

        c1list = [
            'h1bal[1]',
            'h1bal[2]',
            'h1bal[3]',
            'h1bal[4]',
            'h1bal[5]',
            'h2bal[1]',
            'h2bal[2]',
            'h2bal[3]',
            'h2bal[4]',
            'h2bal[5]',
            'F12con[0]',
            'F12con[1]',
            'F12con[2]',
            'F12con[3]',
            'F12con[4]',
            'F12con[5]',
            'Focon[0]',
            'Focon[1]',
            'Focon[2]',
            'Focon[3]',
            'Focon[4]',
            'Focon[5]',
            'min_inflow[1]',
            'min_inflow[2]',
            'min_inflow[3]',
            'min_inflow[4]',
            'min_inflow[5]',
            'max_outflow[0]',
            'max_outflow[1]',
            'max_outflow[2]',
            'max_outflow[3]',
            'max_outflow[4]',
            'max_outflow[5]',
            'h10',
            'h20',
        ]
        c2list = [
            'egb.h1bal_1',
            'egb.h1bal_2',
            'egb.h1bal_3',
            'egb.h1bal_4',
            'egb.h1bal_5',
            'egb.h2bal_1',
            'egb.h2bal_2',
            'egb.h2bal_3',
            'egb.h2bal_4',
            'egb.h2bal_5',
            'egb.output_constraints[F12_0]',
            'egb.output_constraints[F12_1]',
            'egb.output_constraints[F12_2]',
            'egb.output_constraints[F12_3]',
            'egb.output_constraints[F12_4]',
            'egb.output_constraints[F12_5]',
            'egb.output_constraints[Fo_0]',
            'egb.output_constraints[Fo_1]',
            'egb.output_constraints[Fo_2]',
            'egb.output_constraints[Fo_3]',
            'egb.output_constraints[Fo_4]',
            'egb.output_constraints[Fo_5]',
            'min_inflow[1]',
            'min_inflow[2]',
            'min_inflow[3]',
            'min_inflow[4]',
            'min_inflow[5]',
            'max_outflow[0]',
            'max_outflow[1]',
            'max_outflow[2]',
            'max_outflow[3]',
            'max_outflow[4]',
            'max_outflow[5]',
            'h10',
            'h20',
        ]
        c1_c2_map = dict(zip(c1list, c2list))
        c1idx_c2idx_map = {
            i: mex_c_order.index(c1_c2_map[m_c_order[i]]) for i in range(len(m_c_order))
        }

        # get the primals from m and put them in the correct order for mex
        m_x = m_nlp.get_primals()
        mex_x = np.zeros(len(m_x))
        for i in range(len(m_x)):
            mex_x[x1idx_x2idx_map[i]] = m_x[i]

        # get the duals from m and put them in the correct order for mex
        m_lam = m_nlp.get_duals()
        mex_lam = np.zeros(len(m_lam))
        for i in range(len(m_x)):
            mex_lam[c1idx_c2idx_map[i]] = m_lam[i]

        mex_nlp.set_primals(mex_x)
        mex_nlp.set_duals(mex_lam)

        m_obj = m_nlp.evaluate_objective()
        mex_obj = mex_nlp.evaluate_objective()
        self.assertAlmostEqual(m_obj, mex_obj, places=4)

        m_gobj = m_nlp.evaluate_grad_objective()
        mex_gobj = mex_nlp.evaluate_grad_objective()
        check_vectors_specific_order(
            self, m_gobj, m_x_order, mex_gobj, mex_x_order, x1_x2_map
        )

        m_c = m_nlp.evaluate_constraints()
        mex_c = mex_nlp.evaluate_constraints()
        check_vectors_specific_order(
            self, m_c, m_c_order, mex_c, mex_c_order, c1_c2_map
        )

        m_j = m_nlp.evaluate_jacobian()
        mex_j = mex_nlp.evaluate_jacobian().todense()
        check_sparse_matrix_specific_order(
            self,
            m_j,
            m_c_order,
            m_x_order,
            mex_j,
            mex_c_order,
            mex_x_order,
            c1_c2_map,
            x1_x2_map,
        )

        m_h = m_nlp.evaluate_hessian_lag()
        mex_h = mex_nlp.evaluate_hessian_lag()
        check_sparse_matrix_specific_order(
            self,
            m_h,
            m_x_order,
            m_x_order,
            mex_h,
            mex_x_order,
            mex_x_order,
            x1_x2_map,
            x1_x2_map,
        )

        mex_h = 0 * mex_h
        mex_nlp.evaluate_hessian_lag(out=mex_h)
        check_sparse_matrix_specific_order(
            self,
            m_h,
            m_x_order,
            m_x_order,
            mex_h,
            mex_x_order,
            mex_x_order,
            x1_x2_map,
            x1_x2_map,
        )

    @unittest.skipIf(not cyipopt_available, "CyIpopt needed to run tests with solve")
    def test_solve(self):
        A1 = 5
        A2 = 10
        c1 = 3
        c2 = 4
        N = 6
        dt = 1

        m = create_pyomo_model(A1, A2, c1, c2, N, dt)
        solver = pyo.SolverFactory('cyipopt')
        solver.config.options['linear_solver'] = 'mumps'
        status = solver.solve(m, tee=False)

        mex = create_pyomo_external_grey_box_model(A1, A2, c1, c2, N, dt)
        solver = pyo.SolverFactory('cyipopt')
        solver.config.options['linear_solver'] = 'mumps'
        status = solver.solve(mex, tee=False)

        for k in m.F1:
            self.assertAlmostEqual(
                pyo.value(m.F1[k]),
                pyo.value(mex.egb.inputs['F1_{}'.format(k)]),
                places=3,
            )
            self.assertAlmostEqual(
                pyo.value(m.F2[k]),
                pyo.value(mex.egb.inputs['F2_{}'.format(k)]),
                places=3,
            )
        for k in m.h1:
            self.assertAlmostEqual(
                pyo.value(m.h1[k]),
                pyo.value(mex.egb.inputs['h1_{}'.format(k)]),
                places=3,
            )
            self.assertAlmostEqual(
                pyo.value(m.h2[k]),
                pyo.value(mex.egb.inputs['h2_{}'.format(k)]),
                places=3,
            )
        for k in m.F12:
            self.assertAlmostEqual(
                pyo.value(m.F12[k]),
                pyo.value(mex.egb.outputs['F12_{}'.format(k)]),
                places=3,
            )
            self.assertAlmostEqual(
                pyo.value(m.Fo[k]),
                pyo.value(mex.egb.outputs['Fo_{}'.format(k)]),
                places=3,
            )
        """
        self._input_names = ['F1_{}'.format(t) for t in range(1,N)]
        self._input_names.extend(['F2_{}'.format(t) for t in range(1,N)])
        self._input_names.extend(['h1_{}'.format(t) for t in range(0,N)])
        self._input_names.extend(['h2_{}'.format(t) for t in range(0,N)])
        self._output_names = ['F12_{}'.format(t) for t in range(0,N)]
        self._output_names.extend(['Fo_{}'.format(t) for t in range(0,N)])
        """


if __name__ == '__main__':
    t = TestGreyBoxModel()
    t.test_solve()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest
from pyomo.common.dependencies import scipy, scipy_available
import pyomo.environ as pyo

if not scipy_available:
    raise unittest.SkipTest("SciPy is needed to test the SciPy solvers")

from pyomo.contrib.pynumero.asl import AmplInterface
from pyomo.contrib.pynumero.interfaces.pyomo_nlp import PyomoNLP
from pyomo.contrib.pynumero.algorithms.solvers.square_solver_base import (
    SquareNlpSolverBase,
)
from pyomo.contrib.pynumero.algorithms.solvers.scipy_solvers import (
    FsolveNlpSolver,
    RootNlpSolver,
    PyomoScipySolver,
    SecantNewtonNlpSolver,
)


def make_simple_model():
    m = pyo.ConcreteModel()
    m.I = pyo.Set(initialize=[1, 2, 3])
    m.x = pyo.Var(m.I, initialize=1.0)
    m.con1 = pyo.Constraint(expr=m.x[1] ** 2 + m.x[2] ** 2 + m.x[3] ** 2 == 1)
    m.con2 = pyo.Constraint(expr=2 * m.x[1] + 3 * m.x[2] - 4 * m.x[3] == 0)
    m.con3 = pyo.Constraint(expr=m.x[1] == 2 * pyo.exp(m.x[2] / m.x[3]))
    m.obj = pyo.Objective(expr=0.0)
    nlp = PyomoNLP(m)
    return m, nlp


def make_scalar_model():
    m = pyo.ConcreteModel()
    m.x = pyo.Var(initialize=1.0, bounds=(0.0, None))
    m.con = pyo.Constraint(expr=(m.x - 2) ** 3 - 5 * m.x == 0)
    m.obj = pyo.Objective(expr=0.0)
    nlp = PyomoNLP(m)
    return m, nlp


def make_linear_scalar_model():
    m = pyo.ConcreteModel()
    m.x = pyo.Var(initialize=1.0, bounds=(0.0, None))
    m.con = pyo.Constraint(expr=-12.5 * m.x + 30.1 == 0)
    m.obj = pyo.Objective(expr=0.0)
    nlp = PyomoNLP(m)
    return m, nlp


@unittest.skipUnless(AmplInterface.available(), "AmplInterface is not available")
class TestSquareSolverBase(unittest.TestCase):
    def test_not_implemented_solve(self):
        m, nlp = make_simple_model()
        solver = SquareNlpSolverBase(nlp)
        msg = "has not implemented the solve method"
        with self.assertRaisesRegex(NotImplementedError, msg):
            solver.solve()

    def test_not_square(self):
        m, _ = make_simple_model()
        m.con4 = pyo.Constraint(expr=m.x[1] == m.x[2])
        nlp = PyomoNLP(m)
        msg = "same numbers of variables as equality constraints"
        with self.assertRaisesRegex(RuntimeError, msg):
            solver = SquareNlpSolverBase(nlp)

    def test_bounds_and_ineq_okay(self):
        m, _ = make_simple_model()
        m.x[1].setlb(0.0)
        m.x[1].setub(1.0)
        m.con4 = pyo.Constraint(expr=m.x[1] <= m.x[2])
        nlp = PyomoNLP(m)
        # Just construct the solver and get no error
        solver = SquareNlpSolverBase(nlp)


@unittest.skipUnless(AmplInterface.available(), "AmplInterface is not available")
class TestFsolveNLP(unittest.TestCase):
    def test_solve_simple_nlp(self):
        m, nlp = make_simple_model()
        solver = FsolveNlpSolver(nlp, options=dict(xtol=1e-9, maxfev=20, tol=1e-8))
        x, info, ier, msg = solver.solve()
        self.assertEqual(ier, 1)

        variables = [m.x[1], m.x[2], m.x[3]]
        predicted_xorder = [0.92846891, -0.22610731, 0.29465397]
        indices = nlp.get_primal_indices(variables)
        nlp_to_x_indices = [None] * len(variables)
        for i, j in enumerate(indices):
            nlp_to_x_indices[j] = i
        predicted_nlporder = [predicted_xorder[i] for i in nlp_to_x_indices]
        self.assertStructuredAlmostEqual(nlp.get_primals().tolist(), predicted_nlporder)

    def test_solve_max_iter(self):
        m, nlp = make_simple_model()
        solver = FsolveNlpSolver(nlp, options=dict(xtol=1e-9, maxfev=10))
        x, info, ier, msg = solver.solve()
        self.assertNotEqual(ier, 1)
        self.assertIn("has reached maxfev", msg)

    def test_solve_too_tight_tol(self):
        m, nlp = make_simple_model()
        solver = FsolveNlpSolver(nlp, options=dict(xtol=1e-3, maxfev=20, tol=1e-8))
        msg = "does not satisfy the function tolerance"
        with self.assertRaisesRegex(RuntimeError, msg):
            x, info, ier, msg = solver.solve()


@unittest.skipUnless(AmplInterface.available(), "AmplInterface is not available")
class TestPyomoScipySolver(unittest.TestCase):
    def test_available_and_version(self):
        solver = PyomoScipySolver()
        self.assertTrue(solver.available())
        self.assertTrue(solver.license_is_valid())

        sp_version = tuple(int(num) for num in scipy.__version__.split('.'))
        self.assertEqual(sp_version, solver.version())


@unittest.skipUnless(AmplInterface.available(), "AmplInterface is not available")
class TestFsolvePyomo(unittest.TestCase):
    def test_available_and_version(self):
        solver = pyo.SolverFactory("scipy.fsolve")
        self.assertTrue(solver.available())
        self.assertTrue(solver.license_is_valid())

        sp_version = tuple(int(num) for num in scipy.__version__.split('.'))
        self.assertEqual(sp_version, solver.version())

    def test_solve_simple_nlp(self):
        m, _ = make_simple_model()
        solver = pyo.SolverFactory("scipy.fsolve")

        # Just want to make sure this option works
        solver.set_options(dict(full_output=False))

        results = solver.solve(m)
        solution = [m.x[1].value, m.x[2].value, m.x[3].value]
        predicted = [0.92846891, -0.22610731, 0.29465397]
        self.assertStructuredAlmostEqual(solution, predicted)

    def test_solve_results_obj(self):
        m, _ = make_simple_model()
        solver = pyo.SolverFactory("scipy.fsolve")
        results = solver.solve(m)
        solution = [m.x[1].value, m.x[2].value, m.x[3].value]
        predicted = [0.92846891, -0.22610731, 0.29465397]
        self.assertStructuredAlmostEqual(solution, predicted)

        self.assertEqual(results.problem.number_of_constraints, 3)
        self.assertEqual(results.problem.number_of_variables, 3)

        # Note that the solver returns termination condition feasible
        # rather than optimal...
        self.assertEqual(
            results.solver.termination_condition, pyo.TerminationCondition.feasible
        )
        msg = "Solver failed to return an optimal solution"
        with self.assertRaisesRegex(RuntimeError, msg):
            pyo.assert_optimal_termination(results)
        self.assertEqual(results.solver.status, pyo.SolverStatus.ok)

    def test_solve_max_iter(self):
        m, _ = make_simple_model()
        solver = pyo.SolverFactory("scipy.fsolve")
        solver.set_options(dict(xtol=1e-9, maxfev=10))
        res = solver.solve(m)
        self.assertNotEqual(res.solver.return_code, 1)
        self.assertIn("has reached maxfev", res.solver.message)

    def test_solve_too_tight_tol(self):
        m, _ = make_simple_model()
        solver = pyo.SolverFactory(
            "scipy.fsolve", options=dict(xtol=1e-3, maxfev=20, tol=1e-8)
        )
        msg = "does not satisfy the function tolerance"
        with self.assertRaisesRegex(RuntimeError, msg):
            res = solver.solve(m)

    def test_with_scalar_model_bad_starting_point(self):
        # NOTE: fsolve fails to solve this very simple scalar-valued
        # equation with a default starting point (x=1). This may be
        # worth looking into.
        m, _ = make_scalar_model()
        solver = pyo.SolverFactory("scipy.fsolve")
        res = solver.solve(m)
        predicted_x = 4.90547401
        self.assertNotEqual(predicted_x, m.x.value)

    def test_with_scalar_model_good_starting_point(self):
        # NOTE: fsolve can solve this equation with a good starting point.
        m, _ = make_scalar_model()
        m.x.set_value(4.0)
        solver = pyo.SolverFactory("scipy.fsolve")
        res = solver.solve(m)
        predicted_x = 4.90547401
        self.assertAlmostEqual(predicted_x, m.x.value)


@unittest.skipUnless(AmplInterface.available(), "AmplInterface is not available")
class TestRootNLP(unittest.TestCase):
    def test_solve_simple_nlp(self):
        m, nlp = make_simple_model()
        solver = RootNlpSolver(nlp)
        results = solver.solve()
        self.assertTrue(results.success)

        variables = [m.x[1], m.x[2], m.x[3]]
        predicted_xorder = [0.92846891, -0.22610731, 0.29465397]
        indices = nlp.get_primal_indices(variables)
        nlp_to_x_indices = [None] * len(variables)
        for i, j in enumerate(indices):
            nlp_to_x_indices[j] = i
        predicted_nlporder = [predicted_xorder[i] for i in nlp_to_x_indices]
        self.assertStructuredAlmostEqual(results.x.tolist(), predicted_nlporder)

    def test_solve_simple_nlp_levenberg_marquardt(self):
        m, nlp = make_simple_model()
        solver = RootNlpSolver(nlp, options=dict(method="lm"))
        results = solver.solve()
        self.assertTrue(results.success)

        variables = [m.x[1], m.x[2], m.x[3]]
        predicted_xorder = [0.92846891, -0.22610731, 0.29465397]
        indices = nlp.get_primal_indices(variables)
        nlp_to_x_indices = [None] * len(variables)
        for i, j in enumerate(indices):
            nlp_to_x_indices[j] = i
        predicted_nlporder = [predicted_xorder[i] for i in nlp_to_x_indices]
        self.assertStructuredAlmostEqual(results.x.tolist(), predicted_nlporder)


@unittest.skipUnless(AmplInterface.available(), "AmplInterface is not available")
class TestRootPyomo(unittest.TestCase):
    def test_available_and_version(self):
        solver = pyo.SolverFactory("scipy.root")
        self.assertTrue(solver.available())
        self.assertTrue(solver.license_is_valid())

        sp_version = tuple(int(num) for num in scipy.__version__.split('.'))
        self.assertEqual(sp_version, solver.version())

    def test_solve_simple_nlp(self):
        m, _ = make_simple_model()
        solver = pyo.SolverFactory("scipy.root")

        solver.set_options(dict(tol=1e-7))

        results = solver.solve(m)
        solution = [m.x[1].value, m.x[2].value, m.x[3].value]
        predicted = [0.92846891, -0.22610731, 0.29465397]
        self.assertStructuredAlmostEqual(solution, predicted)

    def test_solve_simple_nlp_levenberg_marquardt(self):
        m, _ = make_simple_model()
        solver = pyo.SolverFactory("scipy.root")

        solver.set_options(dict(tol=1e-7, method="lm"))

        results = solver.solve(m)
        solution = [m.x[1].value, m.x[2].value, m.x[3].value]
        predicted = [0.92846891, -0.22610731, 0.29465397]
        self.assertStructuredAlmostEqual(solution, predicted)

    def test_bad_method(self):
        m, _ = make_simple_model()
        solver = pyo.SolverFactory("scipy.root")

        solver.set_options(dict(tol=1e-7, method="some-solver"))
        with self.assertRaisesRegex(ValueError, "not in domain"):
            results = solver.solve(m)

    def test_solver_results_obj(self):
        m, _ = make_simple_model()
        solver = pyo.SolverFactory("scipy.root")

        solver.set_options(dict(tol=1e-7))

        results = solver.solve(m)
        solution = [m.x[1].value, m.x[2].value, m.x[3].value]
        predicted = [0.92846891, -0.22610731, 0.29465397]
        self.assertStructuredAlmostEqual(solution, predicted)

        self.assertEqual(results.problem.number_of_constraints, 3)
        self.assertEqual(results.problem.number_of_variables, 3)
        self.assertEqual(results.solver.return_code, 1)
        self.assertEqual(
            results.solver.termination_condition, pyo.TerminationCondition.feasible
        )
        self.assertEqual(results.solver.message, "The solution converged.")

    def test_solver_results_obj_levenberg_marquardt(self):
        m, _ = make_simple_model()
        solver = pyo.SolverFactory("scipy.root")

        solver.set_options(dict(tol=1e-7, method="lm"))

        results = solver.solve(m)
        solution = [m.x[1].value, m.x[2].value, m.x[3].value]
        predicted = [0.92846891, -0.22610731, 0.29465397]
        self.assertStructuredAlmostEqual(solution, predicted)

        self.assertEqual(results.problem.number_of_constraints, 3)
        self.assertEqual(results.problem.number_of_variables, 3)

        # NOTE: Return code (the scipy OptimizeResult.status field) is not
        # documented in SciPy 1.9.3, so we cannot assert anything about it.
        # self.assertEqual(results.solver.return_code, 1)

        self.assertEqual(
            results.solver.termination_condition, pyo.TerminationCondition.feasible
        )
        self.assertIn(
            "The relative error between two consecutive iterates",
            results.solver.message,
        )


@unittest.skipUnless(AmplInterface.available(), "AmplInterface is not available")
class TestNewtonPyomo(unittest.TestCase):
    def test_available(self):
        solver = pyo.SolverFactory("scipy.newton")
        self.assertTrue(solver.available())
        self.assertTrue(solver.license_is_valid())

        sp_version = tuple(int(num) for num in scipy.__version__.split('.'))
        self.assertEqual(sp_version, solver.version())

    def test_solve(self):
        m, _ = make_scalar_model()
        solver = pyo.SolverFactory("scipy.newton")
        results = solver.solve(m, tee=True)
        predicted_x = 4.90547401
        self.assertAlmostEqual(predicted_x, m.x.value)

    def test_solve_doesnt_converge(self):
        m, _ = make_scalar_model()
        m.x.set_value(3e10)
        solver = pyo.SolverFactory("scipy.newton")
        with self.assertRaisesRegex(RuntimeError, "Failed to converge"):
            # scipy.optimize.newton raises a RuntimeError when it fails to
            # converge to a solution (contrary to fsolve, which happily
            # returns the result). This behavior makes it hard to test
            # for cases where TerminationCondition is not feasible.
            # Should the underlying scipy.optimize.newton call be wrapped
            # with try/except to catch this case and return an infeasible
            # TerminationCondition?
            results = solver.solve(m)

    def test_too_many_iter(self):
        m, _ = make_scalar_model()
        solver = pyo.SolverFactory("scipy.newton")
        solver.set_options({"maxiter": 5})
        with self.assertRaisesRegex(RuntimeError, "Failed to converge"):
            results = solver.solve(m)

    def test_results_object(self):
        m, _ = make_scalar_model()
        solver = pyo.SolverFactory("scipy.newton")
        results = solver.solve(m)
        predicted_x = 4.90547401
        self.assertAlmostEqual(predicted_x, m.x.value)

        # Check results.problem
        self.assertEqual(results.problem.number_of_constraints, 1)
        self.assertEqual(results.problem.number_of_variables, 1)
        self.assertEqual(results.problem.number_of_continuous_variables, 1)
        self.assertEqual(results.problem.number_of_binary_variables, 0)
        self.assertEqual(results.problem.number_of_integer_variables, 0)

        # Assert some reasonable things about the returned results
        self.assertGreater(results.solver.wallclock_time, 0.0)
        self.assertEqual(
            results.solver.termination_condition, pyo.TerminationCondition.feasible
        )
        self.assertEqual(results.solver.status, pyo.SolverStatus.ok)
        self.assertGreater(results.solver.number_of_function_evaluations, 0)

    def test_results_object_without_full_output(self):
        m, _ = make_scalar_model()
        solver = pyo.SolverFactory("scipy.newton")
        solver.set_options(dict(full_output=False))

        results = solver.solve(m)
        predicted_x = 4.90547401
        self.assertAlmostEqual(predicted_x, m.x.value)

        # Check results.problem
        self.assertEqual(results.problem.number_of_constraints, 1)
        self.assertEqual(results.problem.number_of_variables, 1)
        self.assertEqual(results.problem.number_of_continuous_variables, 1)
        self.assertEqual(results.problem.number_of_binary_variables, 0)
        self.assertEqual(results.problem.number_of_integer_variables, 0)

        # Assert some reasonable things about the returned results
        self.assertGreater(results.solver.wallclock_time, 0.0)

        # Now assert that termination condition and solver status have
        # not been reported.
        #
        # This will break if Pyomo changes its default behavior.
        self.assertIs(
            results.solver.termination_condition, pyo.TerminationCondition.unknown
        )
        # The default SolverStatus appears to be ok...
        # self.assertIsNot(results.solver.status, pyo.SolverStatus.ok)

        with self.assertRaises(AttributeError):
            # This attribute has no default, I guess.
            # Assert that it hasn't been set.
            n_eval = results.solver.number_of_function_evaluations


@unittest.skipUnless(AmplInterface.available(), "AmplInterface is not available")
class TestSecantNewton(unittest.TestCase):
    def test_inherited_options_skipped(self):
        m, nlp = make_scalar_model()
        options = SecantNewtonNlpSolver.OPTIONS
        self.assertNotIn("maxiter", options)
        self.assertNotIn("secant", options)
        self.assertIn("secant_iter", options)
        self.assertIn("newton_iter", options)

        with self.assertRaisesRegex(ValueError, "implicit.*keys are not allowed"):
            solver = SecantNewtonNlpSolver(nlp, options={"maxiter": 10})


@unittest.skipUnless(AmplInterface.available(), "AmplInterface is not available")
class TestSecantNewtonPyomo(unittest.TestCase):
    def test_available(self):
        solver = pyo.SolverFactory("scipy.secant-newton")
        self.assertTrue(solver.available())
        self.assertTrue(solver.license_is_valid())

        sp_version = tuple(int(num) for num in scipy.__version__.split('.'))
        self.assertEqual(sp_version, solver.version())

    def test_solve(self):
        m, _ = make_scalar_model()
        solver = pyo.SolverFactory("scipy.secant-newton")
        results = solver.solve(m)
        predicted_x = 4.90547401
        self.assertAlmostEqual(predicted_x, m.x.value)

        self.assertFalse(solver.converged_with_secant())

    def test_solve_doesnt_converge(self):
        m, _ = make_scalar_model()
        m.x.set_value(3e10)
        solver = pyo.SolverFactory("scipy.secant-newton")
        with self.assertRaisesRegex(RuntimeError, "Failed to converge"):
            # scipy.optimize.newton raises a RuntimeError when it fails to
            # converge to a solution (contrary to fsolve, which happily
            # returns the result). This behavior makes it hard to test
            # for cases where TerminationCondition is not feasible.
            # Should the underlying scipy.optimize.newton call be wrapped
            # with try/except to catch this case and return an infeasible
            # TerminationCondition?
            results = solver.solve(m)

        self.assertFalse(solver.converged_with_secant())

    def test_too_many_iter(self):
        m, _ = make_scalar_model()
        solver = pyo.SolverFactory("scipy.secant-newton")

        # We now have a different API (from SolverFactory("newton"),
        # in that we specify the number of iterations for the newton
        # or secant sub-solvers.
        solver.set_options({"newton_iter": 5})
        with self.assertRaisesRegex(RuntimeError, "Failed to converge"):
            results = solver.solve(m)

    def test_results_object(self):
        m, _ = make_scalar_model()
        solver = pyo.SolverFactory("scipy.secant-newton")
        results = solver.solve(m)
        predicted_x = 4.90547401
        self.assertAlmostEqual(predicted_x, m.x.value)

        # Check results.problem
        self.assertEqual(results.problem.number_of_constraints, 1)
        self.assertEqual(results.problem.number_of_variables, 1)
        self.assertEqual(results.problem.number_of_continuous_variables, 1)
        self.assertEqual(results.problem.number_of_binary_variables, 0)
        self.assertEqual(results.problem.number_of_integer_variables, 0)

        # Assert some reasonable things about the returned results
        self.assertGreater(results.solver.wallclock_time, 0.0)
        self.assertEqual(
            results.solver.termination_condition, pyo.TerminationCondition.feasible
        )
        self.assertEqual(results.solver.status, pyo.SolverStatus.ok)
        self.assertGreater(results.solver.number_of_function_evaluations, 0)

    def test_results_object_without_full_output(self):
        m, _ = make_scalar_model()
        solver = pyo.SolverFactory("scipy.secant-newton")
        solver.set_options(dict(full_output=False))

        results = solver.solve(m)
        predicted_x = 4.90547401
        self.assertAlmostEqual(predicted_x, m.x.value)

        # Check results.problem
        self.assertEqual(results.problem.number_of_constraints, 1)
        self.assertEqual(results.problem.number_of_variables, 1)
        self.assertEqual(results.problem.number_of_continuous_variables, 1)
        self.assertEqual(results.problem.number_of_binary_variables, 0)
        self.assertEqual(results.problem.number_of_integer_variables, 0)

        # Assert some reasonable things about the returned results
        self.assertGreater(results.solver.wallclock_time, 0.0)

        # Now assert that termination condition and solver status have
        # not been reported.
        #
        # This will break if Pyomo changes its default behavior.
        self.assertIs(
            results.solver.termination_condition, pyo.TerminationCondition.unknown
        )
        # The default SolverStatus appears to be ok...
        # self.assertIsNot(results.solver.status, pyo.SolverStatus.ok)

        with self.assertRaises(AttributeError):
            # This attribute has no default, I guess.
            # Assert that it hasn't been set.
            n_eval = results.solver.number_of_function_evaluations

    def test_solve_linear(self):
        m, _ = make_linear_scalar_model()
        solver = pyo.SolverFactory("scipy.secant-newton")
        results = solver.solve(m)
        self.assertAlmostEqual(m.x.value, 30.1 / 12.5)
        # This linear equation converges with the secant subsolver.
        self.assertTrue(solver.converged_with_secant())


if __name__ == "__main__":
    unittest.main()
